@@ -35,8 +35,9 @@
  *	Kristian Høgsberg <krh@redhat.com>
  */
 
-#include <errno.h>
 #include "cairoint.h"
+
+#include <errno.h>
 #include <png.h>
 
 /* Unpremultiplies data and converts native endian ARGB => RGBA bytes */
@@ -82,6 +83,25 @@ convert_data_to_bytes (png_structp png, png_row_infop row_info, png_bytep data)
     }
 }
 
+/* Use a couple of simple error callbacks that do not print anything to
+ * stderr and rely on the user to check for errors via the cairo_status_t
+ * return.
+ */
+static void
+png_simple_error_callback (png_structp png_save_ptr,
+	                   png_const_charp error_msg)
+{
+    _cairo_error (CAIRO_STATUS_NO_MEMORY);
+    longjmp (png_save_ptr->jmpbuf, CAIRO_STATUS_NO_MEMORY);
+}
+
+static void
+png_simple_warning_callback (png_structp png_save_ptr,
+	                     png_const_charp error_msg)
+{
+}
+
+
 static cairo_status_t
 write_png (cairo_surface_t	*surface,
 	   png_rw_ptr		write_func,
@@ -117,7 +137,9 @@ write_png (cairo_surface_t	*surface,
     for (i = 0; i < image->height; i++)
 	rows[i] = (png_byte *) image->data + i * image->stride;
 
-    png = png_create_write_struct (PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
+    png = png_create_write_struct (PNG_LIBPNG_VER_STRING, NULL,
+	                           png_simple_error_callback,
+	                           png_simple_warning_callback);
     if (png == NULL) {
 	status = CAIRO_STATUS_NO_MEMORY;
 	goto BAIL2;
@@ -129,10 +151,9 @@ write_png (cairo_surface_t	*surface,
 	goto BAIL3;
     }
 
-    if (setjmp (png_jmpbuf (png))) {
-	status = CAIRO_STATUS_NO_MEMORY;
+    status = setjmp (png_jmpbuf (png));
+    if (status)
 	goto BAIL3;
-    }
 
     png_set_write_fn (png, closure, write_func, NULL);
 
@@ -206,8 +227,13 @@ stdio_write_func (png_structp png, png_bytep data, png_size_t size)
     FILE *fp;
 
     fp = png_get_io_ptr (png);
-    if (fwrite (data, 1, size, fp) != size)
-	png_error(png, "Write Error");
+    while (size) {
+	size_t ret = fwrite (data, 1, size, fp);
+	size -= ret;
+	data += ret;
+	if (size && ferror (fp))
+	    png_error(png, "Write Error");
+    }
 }
 
 /**
@@ -289,7 +315,7 @@ cairo_surface_write_to_png_stream (cairo_surface_t	*surface,
 }
 slim_hidden_def (cairo_surface_write_to_png_stream);
 
-static INLINE int
+static inline int
 multiply_alpha (int alpha, int color)
 {
     int temp = (alpha * color) + 0x80;
@@ -332,20 +358,20 @@ read_png (png_rw_ptr	read_func,
 	  void		*closure)
 {
     cairo_surface_t *surface = (cairo_surface_t*) &_cairo_surface_nil;
-    png_byte *data = NULL;
-    unsigned int i;
     png_struct *png = NULL;
     png_info *info;
+    png_byte *data = NULL;
+    png_byte **row_pointers = NULL;
     png_uint_32 png_width, png_height, stride;
     int depth, color_type, interlace;
+    unsigned int i;
     unsigned int pixel_size;
-    png_byte **row_pointers = NULL;
 
     /* XXX: Perhaps we'll want some other error handlers? */
     png = png_create_read_struct (PNG_LIBPNG_VER_STRING,
                                   NULL,
-                                  NULL,
-                                  NULL);
+	                          png_simple_error_callback,
+	                          png_simple_warning_callback);
     if (png == NULL)
 	goto BAIL;
 
@@ -373,7 +399,11 @@ read_png (png_rw_ptr	read_func,
 
     /* expand gray bit depth if needed */
     if (color_type == PNG_COLOR_TYPE_GRAY && depth < 8)
+#if PNG_LIBPNG_VER >= 10209
+        png_set_expand_gray_1_2_4_to_8 (png);
+#else
         png_set_gray_1_2_4_to_8 (png);
+#endif
     /* transform transparency to alpha */
     if (png_get_valid(png, info, PNG_INFO_tRNS))
         png_set_tRNS_to_alpha (png);
@@ -442,8 +472,13 @@ stdio_read_func (png_structp png, png_bytep data, png_size_t size)
     FILE *fp;
 
     fp = png_get_io_ptr (png);
-    if (fread (data, 1, size, fp) != size)
-	png_error(png, "Read Error");
+    while (size) {
+	size_t ret = fread (data, 1, size, fp);
+	size -= ret;
+	data += ret;
+	if (size && ferror (fp))
+	    png_error(png, "Read Error");
+    }
 }
 
 /**
