@@ -35,89 +35,50 @@
  *	Carl D. Worth <cworth@cworth.org>
  */
 
-#include <stdlib.h>
-
 #include "cairoint.h"
+
 #include "cairo-surface-fallback-private.h"
 #include "cairo-clip-private.h"
 
-const cairo_surface_t _cairo_surface_nil = {
-    &cairo_image_surface_backend,	/* backend */
-    CAIRO_CONTENT_COLOR,
-    CAIRO_SURFACE_TYPE_IMAGE,
-    -1,					/* ref_count */
-    CAIRO_STATUS_NO_MEMORY,		/* status */
-    FALSE,				/* finished */
-    { 0,	/* size */
-      0,	/* num_elements */
-      0,	/* element_size */
-      NULL,	/* elements */
-    },					/* user_data */
-    { 1.0, 0.0,
-      0.0, 1.0,
-      0.0, 0.0
-    },					/* device_transform */
-    { 1.0, 0.0,
-      0.0, 1.0,
-      0.0, 0.0
-    },					/* device_transform_inverse */
-    0.0,				/* x_fallback_resolution */
-    0.0,				/* y_fallback_resolution */
-    0,					/* next_clip_serial */
-    0					/* current_clip_serial */
-};
+#define DEFINE_NIL_SURFACE(status, name)			\
+const cairo_surface_t name = {					\
+    &cairo_image_surface_backend,	/* backend */		\
+    CAIRO_SURFACE_TYPE_IMAGE,					\
+    CAIRO_CONTENT_COLOR,					\
+    CAIRO_REF_COUNT_INVALID,		/* ref_count */		\
+    status,				/* status */		\
+    FALSE,				/* finished */		\
+    { 0,	/* size */					\
+      0,	/* num_elements */				\
+      0,	/* element_size */				\
+      NULL,	/* elements */					\
+    },					/* user_data */		\
+    { 1.0, 0.0,							\
+      0.0, 1.0,							\
+      0.0, 0.0							\
+    },					/* device_transform */	\
+    { 1.0, 0.0,							\
+      0.0, 1.0,							\
+      0.0, 0.0							\
+    },					/* device_transform_inverse */	\
+    0.0,				/* x_fallback_resolution */	\
+    0.0,				/* y_fallback_resolution */	\
+    NULL,				/* clip */		\
+    0,					/* next_clip_serial */	\
+    0,					/* current_clip_serial */	\
+    FALSE,				/* is_snapshot */	\
+    FALSE,				/* has_font_options */	\
+    { CAIRO_ANTIALIAS_DEFAULT,					\
+      CAIRO_SUBPIXEL_ORDER_DEFAULT,				\
+      CAIRO_HINT_STYLE_DEFAULT,					\
+      CAIRO_HINT_METRICS_DEFAULT				\
+    }					/* font_options */	\
+}
 
-const cairo_surface_t _cairo_surface_nil_file_not_found = {
-    &cairo_image_surface_backend,	/* backend */
-    CAIRO_CONTENT_COLOR,
-    CAIRO_SURFACE_TYPE_IMAGE,
-    -1,					/* ref_count */
-    CAIRO_STATUS_FILE_NOT_FOUND,	/* status */
-    FALSE,				/* finished */
-    { 0,	/* size */
-      0,	/* num_elements */
-      0,	/* element_size */
-      NULL,	/* elements */
-    },					/* user_data */
-    { 1.0, 0.0,
-      0.0, 1.0,
-      0.0, 0.0
-    },					/* device_transform */
-    { 1.0, 0.0,
-      0.0, 1.0,
-      0.0, 0.0
-    },					/* device_transform_inverse */
-    0.0,				/* x_fallback_resolution */
-    0.0,				/* y_fallback_resolution */
-    0,					/* next_clip_serial */
-    0					/* current_clip_serial */
-};
-
-const cairo_surface_t _cairo_surface_nil_read_error = {
-    &cairo_image_surface_backend,	/* backend */
-    CAIRO_CONTENT_COLOR,
-    CAIRO_SURFACE_TYPE_IMAGE,
-    -1,					/* ref_count */
-    CAIRO_STATUS_READ_ERROR,		/* status */
-    FALSE,				/* finished */
-    { 0,	/* size */
-      0,	/* num_elements */
-      0,	/* element_size */
-      NULL,	/* elements */
-    },					/* user_data */
-    { 1.0, 0.0,
-      0.0, 1.0,
-      0.0, 0.0
-    },					/* device_transform */
-    { 1.0, 0.0,
-      0.0, 1.0,
-      0.0, 0.0
-    },					/* device_transform_inverse */
-    0.0,				/* x_fallback_resolution */
-    0.0,				/* y_fallback_resolution */
-    0,					/* next_clip_serial */
-    0					/* current_clip_serial */
-};
+DEFINE_NIL_SURFACE(CAIRO_STATUS_NO_MEMORY, _cairo_surface_nil);
+DEFINE_NIL_SURFACE(CAIRO_STATUS_FILE_NOT_FOUND, _cairo_surface_nil_file_not_found);
+DEFINE_NIL_SURFACE(CAIRO_STATUS_READ_ERROR, _cairo_surface_nil_read_error);
+DEFINE_NIL_SURFACE(CAIRO_STATUS_WRITE_ERROR, _cairo_surface_nil_write_error);
 
 static void
 _cairo_surface_copy_pattern_for_destination (const cairo_pattern_t *pattern,
@@ -157,7 +118,10 @@ _cairo_surface_set_error (cairo_surface_t *surface,
  * cairo_surface_get_type:
  * @surface: a #cairo_surface_t
  *
- * Return value: The type of @surface. See #cairo_surface_type_t.
+ * This function returns the type of the backend used to create
+ * a surface. See #cairo_surface_type_t for available types.
+ *
+ * Return value: The type of @surface.
  *
  * Since: 1.2
  **/
@@ -170,15 +134,18 @@ cairo_surface_get_type (cairo_surface_t *surface)
      * surface. */
     return surface->type;
 }
+slim_hidden_def (cairo_surface_get_type);
 
 /**
  * cairo_surface_get_content:
  * @surface: a #cairo_surface_t
  *
- * Return value: The content type of @surface which indicates whether
- * the surface contains color and/or alpha information. See
+ * This function returns the content type of @surface which indicates
+ * whether the surface contains color and/or alpha information. See
  * #cairo_content_t.
  *
+ * Return value: The content type of @surface.
+ *
  * Since: 1.2
  **/
 cairo_content_t
@@ -205,16 +172,17 @@ cairo_surface_status (cairo_surface_t *surface)
 {
     return surface->status;
 }
+slim_hidden_def (cairo_surface_status);
 
 void
 _cairo_surface_init (cairo_surface_t			*surface,
 		     const cairo_surface_backend_t	*backend,
 		     cairo_content_t			 content)
 {
+    CAIRO_MUTEX_INITIALIZE ();
+
     surface->backend = backend;
-
     surface->content = content;
-
     surface->type = backend->type;
 
     surface->ref_count = 1;
@@ -234,6 +202,8 @@ _cairo_surface_init (cairo_surface_t			*surface,
     surface->current_clip_serial = 0;
 
     surface->is_snapshot = FALSE;
+
+    surface->has_font_options = FALSE;
 }
 
 cairo_surface_t *
@@ -242,15 +212,39 @@ _cairo_surface_create_similar_scratch (cairo_surface_t *other,
 				       int		width,
 				       int		height)
 {
+    cairo_surface_t *surface = NULL;
+    cairo_font_options_t options;
+
     cairo_format_t format = _cairo_format_from_content (content);
 
     if (other->status)
 	return (cairo_surface_t*) &_cairo_surface_nil;
 
-    if (other->backend->create_similar)
-	return other->backend->create_similar (other, content, width, height);
-    else
-	return cairo_image_surface_create (format, width, height);
+    if (other->backend->create_similar) {
+	surface = other->backend->create_similar (other, content, width, height);
+	/* It's not an error if the backend didn't create a valid
+	 * surface---it may just not be supported. */
+	if (surface && surface->status) {
+	    cairo_surface_destroy (surface);
+	    surface = NULL;
+	}
+    }
+
+    if (surface == NULL)
+	surface = cairo_image_surface_create (format, width, height);
+
+    /* If any error occurred, then return the nil surface we received. */
+    if (surface->status)
+	return surface;
+
+    cairo_surface_get_font_options (other, &options);
+    _cairo_surface_set_font_options (surface, &options);
+
+    cairo_surface_set_fallback_resolution (surface,
+					   other->x_fallback_resolution,
+					   other->y_fallback_resolution);
+
+    return surface;
 }
 
 /**
@@ -261,9 +255,12 @@ _cairo_surface_create_similar_scratch (cairo_surface_t *other,
  * @height: height of the new surface (in device-space units)
  *
  * Create a new surface that is as compatible as possible with an
- * existing surface. The new surface will use the same backend as
- * @other unless that is not possible for some reason. The type of the
- * returned surface may be examined with cairo_surface_get_type().
+ * existing surface. For example the new surface will have the same
+ * fallback resolution and font options as @other. Generally, the new
+ * surface will also use the same backend as @other, unless that is
+ * not possible for some reason. The type of the returned surface may
+ * be examined with cairo_surface_get_type().
+ *
  * Initially the surface contents are all 0 (transparent if contents
  * have transparency, black otherwise.)
  *
@@ -291,15 +288,18 @@ cairo_surface_create_similar (cairo_surface_t  *other,
 
     return _cairo_surface_create_similar_solid (other, content,
 						width, height,
-						CAIRO_COLOR_TRANSPARENT);
+						CAIRO_COLOR_TRANSPARENT,
+						NULL);
 }
+slim_hidden_def (cairo_surface_create_similar);
 
 cairo_surface_t *
 _cairo_surface_create_similar_solid (cairo_surface_t	 *other,
 				     cairo_content_t	  content,
 				     int		  width,
 				     int		  height,
-				     const cairo_color_t *color)
+				     const cairo_color_t *color,
+				     cairo_pattern_t	 *pattern)
 {
     cairo_status_t status;
     cairo_surface_t *surface;
@@ -312,19 +312,23 @@ _cairo_surface_create_similar_solid (cairo_surface_t	 *other,
 	return (cairo_surface_t*) &_cairo_surface_nil;
     }
 
-    source = _cairo_pattern_create_solid (color);
-    if (source->status) {
-	cairo_surface_destroy (surface);
-	_cairo_error (CAIRO_STATUS_NO_MEMORY);
-	return (cairo_surface_t*) &_cairo_surface_nil;
-    }
+    if (pattern == NULL) {
+	source = _cairo_pattern_create_solid (color, content);
+	if (source->status) {
+	    cairo_surface_destroy (surface);
+	    _cairo_error (CAIRO_STATUS_NO_MEMORY);
+	    return (cairo_surface_t*) &_cairo_surface_nil;
+	}
+    } else
+	source = pattern;
 
     status = _cairo_surface_paint (surface,
 				   color == CAIRO_COLOR_TRANSPARENT ?
 				   CAIRO_OPERATOR_CLEAR :
 				   CAIRO_OPERATOR_SOURCE, source);
 
-    cairo_pattern_destroy (source);
+    if (source != pattern)
+	cairo_pattern_destroy (source);
 
     if (status) {
 	cairo_surface_destroy (surface);
@@ -354,15 +358,15 @@ _cairo_surface_get_clip_mode (cairo_surface_t *surface)
  * @surface from being destroyed until a matching call to
  * cairo_surface_destroy() is made.
  *
+ * The number of references to a #cairo_surface_t can be get using
+ * cairo_surface_get_reference_count().
+ *
  * Return value: the referenced #cairo_surface_t.
  **/
 cairo_surface_t *
 cairo_surface_reference (cairo_surface_t *surface)
 {
-    if (surface == NULL)
-	return NULL;
-
-    if (surface->ref_count == (unsigned int)-1)
+    if (surface == NULL || surface->ref_count == CAIRO_REF_COUNT_INVALID)
 	return surface;
 
     assert (surface->ref_count > 0);
@@ -371,10 +375,11 @@ cairo_surface_reference (cairo_surface_t *surface)
 
     return surface;
 }
+slim_hidden_def (cairo_surface_reference);
 
 /**
  * cairo_surface_destroy:
- * @surface: a #cairo_t
+ * @surface: a #cairo_surface_t
  *
  * Decreases the reference count on @surface by one. If the result is
  * zero, then @surface and all associated resources are freed.  See
@@ -383,10 +388,7 @@ cairo_surface_reference (cairo_surface_t *surface)
 void
 cairo_surface_destroy (cairo_surface_t *surface)
 {
-    if (surface == NULL)
-	return;
-
-    if (surface->ref_count == (unsigned int)-1)
+    if (surface == NULL || surface->ref_count == CAIRO_REF_COUNT_INVALID)
 	return;
 
     assert (surface->ref_count > 0);
@@ -395,7 +397,8 @@ cairo_surface_destroy (cairo_surface_t *surface)
     if (surface->ref_count)
 	return;
 
-    cairo_surface_finish (surface);
+    if (! surface->finished)
+	cairo_surface_finish (surface);
 
     _cairo_user_data_array_fini (&surface->user_data);
 
@@ -403,6 +406,26 @@ cairo_surface_destroy (cairo_surface_t *surface)
 }
 slim_hidden_def(cairo_surface_destroy);
 
+/**
+ * cairo_surface_get_reference_count:
+ * @surface: a #cairo_surface_t
+ *
+ * Returns the current reference count of @surface.
+ *
+ * Return value: the current reference count of @surface.  If the
+ * object is a nil object, 0 will be returned.
+ *
+ * Since: 1.4
+ **/
+unsigned int
+cairo_surface_get_reference_count (cairo_surface_t *surface)
+{
+    if (surface == NULL || surface->ref_count == CAIRO_REF_COUNT_INVALID)
+	return 0;
+
+    return surface->ref_count;
+}
+
 /**
  * cairo_surface_finish:
  * @surface: the #cairo_surface_t to finish
@@ -426,6 +449,12 @@ cairo_surface_finish (cairo_surface_t *surface)
 {
     cairo_status_t status;
 
+    if (surface == NULL)
+	return;
+
+    if (surface->ref_count == CAIRO_REF_COUNT_INVALID)
+	return;
+
     if (surface->finished) {
 	_cairo_surface_set_error (surface, CAIRO_STATUS_SURFACE_FINISHED);
 	return;
@@ -452,6 +481,7 @@ cairo_surface_finish (cairo_surface_t *surface)
 
     surface->finished = TRUE;
 }
+slim_hidden_def (cairo_surface_finish);
 
 /**
  * cairo_surface_get_user_data:
@@ -495,13 +525,40 @@ cairo_surface_set_user_data (cairo_surface_t		 *surface,
 			     void			 *user_data,
 			     cairo_destroy_func_t	 destroy)
 {
-    if (surface->ref_count == -1)
+    if (surface->ref_count == CAIRO_REF_COUNT_INVALID)
 	return CAIRO_STATUS_NO_MEMORY;
 
     return _cairo_user_data_array_set_data (&surface->user_data,
 					    key, user_data, destroy);
 }
 
+/**
+ * _cairo_surface_set_font_options:
+ * @surface: a #cairo_surface_t
+ * @options: a #cairo_font_options_t object that contains the
+ *   options to use for this surface instead of backend's default
+ *   font options.
+ *
+ * Sets the default font rendering options for the surface.
+ * This is useful to correctly propagate default font options when
+ * falling back to an image surface in a backend implementation.
+ * This affects the options returned in cairo_surface_get_font_options().
+ *
+ * If @options is %NULL the surface options are reset to those of
+ * the backend default.
+ **/
+void
+_cairo_surface_set_font_options (cairo_surface_t       *surface,
+				 cairo_font_options_t  *options)
+{
+    if (options) {
+	surface->has_font_options = TRUE;
+	_cairo_font_options_init_copy (&surface->font_options, options);
+    } else {
+	surface->has_font_options = FALSE;
+    }
+}
+
 /**
  * cairo_surface_get_font_options:
  * @surface: a #cairo_surface_t
@@ -518,12 +575,19 @@ void
 cairo_surface_get_font_options (cairo_surface_t       *surface,
 				cairo_font_options_t  *options)
 {
-    if (!surface->finished && surface->backend->get_font_options) {
-	surface->backend->get_font_options (surface, options);
-    } else {
-	_cairo_font_options_init_default (options);
+    if (!surface->has_font_options) {
+	surface->has_font_options = TRUE;
+
+	_cairo_font_options_init_default (&surface->font_options);
+
+	if (!surface->finished && surface->backend->get_font_options) {
+	    surface->backend->get_font_options (surface, &surface->font_options);
+	}
     }
+
+    _cairo_font_options_init_copy (options, &surface->font_options);
 }
+slim_hidden_def (cairo_surface_get_font_options);
 
 /**
  * cairo_surface_flush:
@@ -630,6 +694,7 @@ cairo_surface_mark_dirty_rectangle (cairo_surface_t *surface,
 	    _cairo_surface_set_error (surface, status);
     }
 }
+slim_hidden_def (cairo_surface_mark_dirty_rectangle);
 
 /**
  * _cairo_surface_set_device_scale:
@@ -712,6 +777,7 @@ cairo_surface_set_device_offset (cairo_surface_t *surface,
     surface->device_transform_inverse.x0 = - x_offset;
     surface->device_transform_inverse.y0 = - y_offset;
 }
+slim_hidden_def (cairo_surface_set_device_offset);
 
 /**
  * cairo_surface_get_device_offset:
@@ -729,9 +795,12 @@ cairo_surface_get_device_offset (cairo_surface_t *surface,
 				 double          *x_offset,
 				 double          *y_offset)
 {
-    *x_offset = surface->device_transform.x0;
-    *y_offset = surface->device_transform.y0;
+    if (x_offset)
+	*x_offset = surface->device_transform.x0;
+    if (y_offset)
+	*y_offset = surface->device_transform.y0;
 }
+slim_hidden_def (cairo_surface_get_device_offset);
 
 /**
  * cairo_surface_set_fallback_resolution:
@@ -771,6 +840,7 @@ cairo_surface_set_fallback_resolution (cairo_surface_t	*surface,
     surface->x_fallback_resolution = x_pixels_per_inch;
     surface->y_fallback_resolution = y_pixels_per_inch;
 }
+slim_hidden_def (cairo_surface_set_fallback_resolution);
 
 cairo_bool_t
 _cairo_surface_has_device_transform (cairo_surface_t *surface)
@@ -899,6 +969,10 @@ _cairo_surface_release_dest_image (cairo_surface_t         *surface,
  * _cairo_surface_clone_similar:
  * @surface: a #cairo_surface_t
  * @src: the source image
+ * @src_x: extent for the rectangle in src we actually care about
+ * @src_y: extent for the rectangle in src we actually care about
+ * @width: extent for the rectangle in src we actually care about
+ * @height: extent for the rectangle in src we actually care about
  * @clone_out: location to store a surface compatible with @surface
  *   and with contents identical to @src. The caller must call
  *   cairo_surface_destroy() on the result.
@@ -914,6 +988,10 @@ _cairo_surface_release_dest_image (cairo_surface_t         *surface,
 cairo_status_t
 _cairo_surface_clone_similar (cairo_surface_t  *surface,
 			      cairo_surface_t  *src,
+			      int               src_x,
+			      int               src_y,
+			      int               width,
+			      int               height,
 			      cairo_surface_t **clone_out)
 {
     cairo_status_t status;
@@ -926,8 +1004,9 @@ _cairo_surface_clone_similar (cairo_surface_t  *surface,
     if (surface->backend->clone_similar == NULL)
 	return CAIRO_INT_STATUS_UNSUPPORTED;
 
-    status = surface->backend->clone_similar (surface, src, clone_out);
-    if (status == CAIRO_STATUS_SUCCESS)
+    status = surface->backend->clone_similar (surface, src, src_x, src_y,
+					      width, height, clone_out);
+    if (status == CAIRO_STATUS_SUCCESS && *clone_out != src)
         (*clone_out)->device_transform = src->device_transform;
 
     if (status != CAIRO_INT_STATUS_UNSUPPORTED)
@@ -937,9 +1016,12 @@ _cairo_surface_clone_similar (cairo_surface_t  *surface,
     if (status != CAIRO_STATUS_SUCCESS)
 	return status;
 
-    status = surface->backend->clone_similar (surface, &image->base, clone_out);
-    if (status == CAIRO_STATUS_SUCCESS)
+    status = surface->backend->clone_similar (surface, &image->base, src_x,
+					      src_y, width, height, clone_out);
+    if (status == CAIRO_STATUS_SUCCESS && *clone_out != src) {
         (*clone_out)->device_transform = src->device_transform;
+        (*clone_out)->device_transform_inverse = src->device_transform_inverse;
+    }
 
     /* If the above failed point, we could implement a full fallback
      * using acquire_dest_image, but that's going to be very
@@ -1093,6 +1175,7 @@ _cairo_surface_fill_region (cairo_surface_t	   *surface,
 {
     int num_rects = pixman_region_num_rects (region);
     pixman_box16_t *boxes = pixman_region_rects (region);
+    cairo_rectangle_int16_t stack_rects[CAIRO_STACK_BUFFER_SIZE / sizeof (cairo_rectangle_int16_t)];
     cairo_rectangle_int16_t *rects;
     cairo_status_t status;
     int i;
@@ -1102,9 +1185,12 @@ _cairo_surface_fill_region (cairo_surface_t	   *surface,
     if (!num_rects)
 	return CAIRO_STATUS_SUCCESS;
 
-    rects = malloc (sizeof (pixman_rectangle_t) * num_rects);
-    if (!rects)
-	return CAIRO_STATUS_NO_MEMORY;
+    rects = stack_rects;
+    if (num_rects > ARRAY_LENGTH (stack_rects)) {
+	rects = malloc (sizeof (cairo_rectangle_int16_t) * num_rects);
+	if (!rects)
+	    return CAIRO_STATUS_NO_MEMORY;
+    }
 
     for (i = 0; i < num_rects; i++) {
 	rects[i].x = boxes[i].x1;
@@ -1116,7 +1202,8 @@ _cairo_surface_fill_region (cairo_surface_t	   *surface,
     status =  _cairo_surface_fill_rectangles (surface, op,
 					      color, rects, num_rects);
 
-    free (rects);
+    if (rects != stack_rects)
+	free (rects);
 
     return status;
 }
@@ -1178,12 +1265,6 @@ _cairo_surface_paint (cairo_surface_t	*surface,
 
     assert (! surface->is_snapshot);
 
-    if (source->type == CAIRO_PATTERN_TYPE_SURFACE &&
-	(source->extend == CAIRO_EXTEND_REFLECT || source->extend == CAIRO_EXTEND_PAD))
-    {
-        return CAIRO_STATUS_NO_MEMORY;
-    }
-
     _cairo_surface_copy_pattern_for_destination (source, surface, &dev_source.base);
 
     if (surface->backend->paint) {
@@ -1212,12 +1293,6 @@ _cairo_surface_mask (cairo_surface_t	*surface,
 
     assert (! surface->is_snapshot);
 
-    if (source->type == CAIRO_PATTERN_TYPE_SURFACE &&
-	(source->extend == CAIRO_EXTEND_REFLECT || source->extend == CAIRO_EXTEND_PAD))
-    {
-        return CAIRO_STATUS_NO_MEMORY;
-    }
-
     _cairo_surface_copy_pattern_for_destination (source, surface, &dev_source.base);
     _cairo_surface_copy_pattern_for_destination (mask, surface, &dev_mask.base);
 
@@ -1256,12 +1331,6 @@ _cairo_surface_stroke (cairo_surface_t		*surface,
 
     assert (! surface->is_snapshot);
 
-    if (source->type == CAIRO_PATTERN_TYPE_SURFACE &&
-	(source->extend == CAIRO_EXTEND_REFLECT || source->extend == CAIRO_EXTEND_PAD))
-    {
-        return CAIRO_STATUS_NO_MEMORY;
-    }
-
     _cairo_surface_copy_pattern_for_destination (source, surface, &dev_source.base);
 
     if (surface->backend->stroke) {
@@ -1301,12 +1370,6 @@ _cairo_surface_fill (cairo_surface_t	*surface,
 
     assert (! surface->is_snapshot);
 
-    if (source->type == CAIRO_PATTERN_TYPE_SURFACE &&
-	(source->extend == CAIRO_EXTEND_REFLECT || source->extend == CAIRO_EXTEND_PAD))
-    {
-        return CAIRO_STATUS_NO_MEMORY;
-    }
-
     _cairo_surface_copy_pattern_for_destination (source, surface, &dev_source.base);
 
     if (surface->backend->fill) {
@@ -1377,10 +1440,7 @@ _cairo_surface_composite_trapezoids (cairo_operator_t		op,
 							  traps, num_traps);
 }
 
-/* _copy_page and _show_page are unique among _cairo_surface functions
- * in that they will actually return CAIRO_INT_STATUS_UNSUPPORTED
- * rather than performing any fallbacks. */
-cairo_int_status_t
+cairo_status_t
 _cairo_surface_copy_page (cairo_surface_t *surface)
 {
     assert (! surface->is_snapshot);
@@ -1391,16 +1451,14 @@ _cairo_surface_copy_page (cairo_surface_t *surface)
     if (surface->finished)
 	return CAIRO_STATUS_SURFACE_FINISHED;
 
+    /* It's fine if some backends don't implement copy_page */
     if (surface->backend->copy_page == NULL)
-	return CAIRO_INT_STATUS_UNSUPPORTED;
+	return CAIRO_STATUS_SUCCESS;
 
     return surface->backend->copy_page (surface);
 }
 
-/* _show_page and _copy_page are unique among _cairo_surface functions
- * in that they will actually return CAIRO_INT_STATUS_UNSUPPORTED
- * rather than performing any fallbacks. */
-cairo_int_status_t
+cairo_status_t
 _cairo_surface_show_page (cairo_surface_t *surface)
 {
     assert (! surface->is_snapshot);
@@ -1411,8 +1469,9 @@ _cairo_surface_show_page (cairo_surface_t *surface)
     if (surface->finished)
 	return CAIRO_STATUS_SURFACE_FINISHED;
 
+    /* It's fine if some backends don't implement show_page */
     if (surface->backend->show_page == NULL)
-	return CAIRO_INT_STATUS_UNSUPPORTED;
+	return CAIRO_STATUS_SUCCESS;
 
     return surface->backend->show_page (surface);
 }
@@ -1647,9 +1706,9 @@ _cairo_surface_set_clip (cairo_surface_t *surface, cairo_clip_t *clip)
 						 clip->path,
 						 clip->serial);
 
-	if (clip->region)
+	if (clip->has_region)
 	    return _cairo_surface_set_clip_region (surface,
-						   clip->region,
+						   &clip->region,
 						   clip->serial);
     }
 
@@ -1662,7 +1721,7 @@ _cairo_surface_set_clip (cairo_surface_t *surface, cairo_clip_t *clip)
  *
  * This function returns a bounding box for the surface.  The surface
  * bounds are defined as a region beyond which no rendering will
- * possibly be recorded, in otherwords, it is the maximum extent of
+ * possibly be recorded, in other words, it is the maximum extent of
  * potentially usable coordinates.
  *
  * For vector surfaces, (PDF, PS, SVG and meta-surfaces), the surface
@@ -1697,81 +1756,65 @@ cairo_status_t
 _cairo_surface_show_glyphs (cairo_surface_t	*surface,
 			    cairo_operator_t	 op,
 			    cairo_pattern_t	*source,
-			    const cairo_glyph_t	*glyphs,
+			    cairo_glyph_t	*glyphs,
 			    int			 num_glyphs,
 			    cairo_scaled_font_t	*scaled_font)
 {
     cairo_status_t status;
-    cairo_glyph_t *dev_glyphs = (cairo_glyph_t*) glyphs;
     cairo_scaled_font_t *dev_scaled_font = scaled_font;
     cairo_pattern_union_t dev_source;
+    cairo_matrix_t font_matrix;
 
     assert (! surface->is_snapshot);
 
-    if (source->type == CAIRO_PATTERN_TYPE_SURFACE &&
-	(source->extend == CAIRO_EXTEND_REFLECT || source->extend == CAIRO_EXTEND_PAD))
-    {
-        return CAIRO_STATUS_NO_MEMORY;
-    }
+    if (!num_glyphs)
+	return CAIRO_STATUS_SUCCESS;
 
     _cairo_surface_copy_pattern_for_destination (source,
 						 surface,
 						 &dev_source.base);
 
-    if (_cairo_surface_has_device_transform (surface))
+    cairo_scaled_font_get_font_matrix (scaled_font, &font_matrix);
+
+    if (_cairo_surface_has_device_transform (surface) &&
+	! _cairo_matrix_is_integer_translation (&surface->device_transform, NULL, NULL))
     {
-        int i;
+	cairo_font_options_t *font_options;
+	cairo_matrix_t dev_ctm;
 
-        dev_glyphs = malloc (sizeof(cairo_glyph_t) * num_glyphs);
-        if (!dev_glyphs)
-            return CAIRO_STATUS_NO_MEMORY;
+	font_options = cairo_font_options_create ();
 
-        for (i = 0; i < num_glyphs; i++) {
-            dev_glyphs[i].index = glyphs[i].index;
-            dev_glyphs[i].x = glyphs[i].x;
-	    dev_glyphs[i].y = glyphs[i].y;
-	    cairo_matrix_transform_point (&surface->device_transform,
-					  &dev_glyphs[i].x,
-					  &dev_glyphs[i].y);
-        }
-
-	if (! _cairo_matrix_is_integer_translation (&surface->device_transform, NULL, NULL)) {
-	    cairo_font_options_t *font_options;
-	    cairo_matrix_t font_matrix, dev_ctm;
-
-	    font_options = cairo_font_options_create ();
-
-	    cairo_scaled_font_get_font_matrix (scaled_font, &font_matrix);
-	    cairo_scaled_font_get_ctm (scaled_font, &dev_ctm);
-	    cairo_matrix_multiply (&dev_ctm, &dev_ctm, &surface->device_transform);
-	    cairo_scaled_font_get_font_options (scaled_font, font_options);
-	    dev_scaled_font = cairo_scaled_font_create (cairo_scaled_font_get_font_face (scaled_font),
-							&font_matrix,
-							&dev_ctm,
-							font_options);
-	    cairo_font_options_destroy (font_options);
-	}
+	cairo_scaled_font_get_ctm (scaled_font, &dev_ctm);
+	cairo_matrix_multiply (&dev_ctm, &dev_ctm, &surface->device_transform);
+	cairo_scaled_font_get_font_options (scaled_font, font_options);
+	dev_scaled_font = cairo_scaled_font_create (cairo_scaled_font_get_font_face (scaled_font),
+						    &font_matrix,
+						    &dev_ctm,
+						    font_options);
+	cairo_font_options_destroy (font_options);
     }
 
-    if (surface->backend->show_glyphs) {
+    CAIRO_MUTEX_LOCK (dev_scaled_font->mutex);
+
+    status = CAIRO_INT_STATUS_UNSUPPORTED;
+
+    if (surface->backend->show_glyphs)
 	status = surface->backend->show_glyphs (surface, op, &dev_source.base,
-						dev_glyphs, num_glyphs,
+						glyphs, num_glyphs,
                                                 dev_scaled_font);
-	if (status != CAIRO_INT_STATUS_UNSUPPORTED)
-            goto FINISH;
-    }
 
-    status = _cairo_surface_fallback_show_glyphs (surface, op, &dev_source.base,
-                                                  dev_glyphs, num_glyphs,
-                                                  dev_scaled_font);
+    if (status == CAIRO_INT_STATUS_UNSUPPORTED)
+	status = _cairo_surface_fallback_show_glyphs (surface, op, &dev_source.base,
+						      glyphs, num_glyphs,
+						      dev_scaled_font);
 
-FINISH:
-    if (dev_glyphs != glyphs)
-        free (dev_glyphs);
+    CAIRO_MUTEX_UNLOCK (dev_scaled_font->mutex);
 
     if (dev_scaled_font != scaled_font)
 	cairo_scaled_font_destroy (dev_scaled_font);
 
+    _cairo_pattern_fini (&dev_source.base);
+
     return status;
 }
 
@@ -1791,7 +1834,7 @@ _cairo_surface_old_show_glyphs (cairo_scaled_font_t	*scaled_font,
 				int			 dest_y,
 				unsigned int		 width,
 				unsigned int		 height,
-				const cairo_glyph_t	*glyphs,
+				cairo_glyph_t		*glyphs,
 				int			 num_glyphs)
 {
     cairo_status_t status;
@@ -1828,9 +1871,11 @@ _cairo_surface_composite_fixup_unbounded_internal (cairo_surface_t         *dst,
 {
     cairo_rectangle_int16_t dst_rectangle;
     cairo_rectangle_int16_t drawn_rectangle;
-    pixman_region16_t *drawn_region;
-    pixman_region16_t *clear_region;
-    cairo_status_t status = CAIRO_STATUS_SUCCESS;
+    cairo_bool_t has_drawn_region = FALSE;
+    cairo_bool_t has_clear_region = FALSE;
+    pixman_region16_t drawn_region;
+    pixman_region16_t clear_region;
+    cairo_status_t status;
 
     /* The area that was drawn is the area in the destination rectangle but not within
      * the source or the mask.
@@ -1843,34 +1888,38 @@ _cairo_surface_composite_fixup_unbounded_internal (cairo_surface_t         *dst,
     drawn_rectangle = dst_rectangle;
 
     if (src_rectangle)
-	_cairo_rectangle_intersect (&drawn_rectangle, src_rectangle);
+        _cairo_rectangle_intersect (&drawn_rectangle, src_rectangle);
 
     if (mask_rectangle)
-	_cairo_rectangle_intersect (&drawn_rectangle, mask_rectangle);
+        _cairo_rectangle_intersect (&drawn_rectangle, mask_rectangle);
 
     /* Now compute the area that is in dst_rectangle but not in drawn_rectangle
      */
-    drawn_region = _cairo_region_create_from_rectangle (&drawn_rectangle);
-    clear_region = _cairo_region_create_from_rectangle (&dst_rectangle);
-    if (!drawn_region || !clear_region) {
-	status = CAIRO_STATUS_NO_MEMORY;
-	goto CLEANUP_REGIONS;
-    }
+    pixman_region_init_rect (&drawn_region,
+                              drawn_rectangle.x, drawn_rectangle.y,
+                              drawn_rectangle.width, drawn_rectangle.height);
+    pixman_region_init_rect (&clear_region,
+                              dst_rectangle.x, dst_rectangle.y,
+                              dst_rectangle.width, dst_rectangle.height);
 
-    if (pixman_region_subtract (clear_region, clear_region, drawn_region) != PIXMAN_REGION_STATUS_SUCCESS) {
-	status = CAIRO_STATUS_NO_MEMORY;
-	goto CLEANUP_REGIONS;
+    has_drawn_region = TRUE;
+    has_clear_region = TRUE;
+
+    if (PIXMAN_REGION_STATUS_SUCCESS !=
+        pixman_region_subtract (&clear_region, &clear_region, &drawn_region)) {
+        status = CAIRO_STATUS_NO_MEMORY;
+        goto CLEANUP_REGIONS;
     }
 
     status = _cairo_surface_fill_region (dst, CAIRO_OPERATOR_SOURCE,
-					 CAIRO_COLOR_TRANSPARENT,
-					 clear_region);
+                                         CAIRO_COLOR_TRANSPARENT,
+                                         &clear_region);
 
- CLEANUP_REGIONS:
-    if (drawn_region)
-	pixman_region_destroy (drawn_region);
-    if (clear_region)
-	pixman_region_destroy (clear_region);
+CLEANUP_REGIONS:
+    if (has_drawn_region)
+        pixman_region_fini (&drawn_region);
+    if (has_clear_region)
+        pixman_region_fini (&clear_region);
 
     return status;
 }
