@@ -1,6 +1,6 @@
 /*
  * Copyright (c) 2004 Ximian
- * Copyright (c) 2004-2006 Novell, Inc.
+ * Copyright (c) 2004-2007 Novell, Inc.
  * 
  * Permission is hereby granted, free of charge, to any person obtaining a copy of this software 
  * and associated documentation files (the "Software"), to deal in the Software without restriction, 
@@ -17,17 +17,17 @@
  * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
  * OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
  * 
- * Author:
- *          Jordi Mas i Hernandez <jordi@ximian.com>, 2004-2006
- *          Peter Dennis Bartok <pbartok@novell.com>
+ * Authors:
+ *	Jordi Mas i Hernandez <jordi@ximian.com>, 2004-2006
+ *	Peter Dennis Bartok <pbartok@novell.com>
+ *	Sebastien Pouliot  <sebastien@ximian.com>
  */
-#include <stdio.h>
-#include "gdip.h"
-#include <math.h>
-#include <glib.h>
-#include <freetype/tttables.h>
-#include <pthread.h>
 
+#include "gdiplus-private.h"
+#include "font-private.h"
+#include "fontcollection-private.h"
+#include "fontfamily-private.h"
+#include "graphics-private.h"
 
 /* Generic fonts families */
 static GStaticMutex generic = G_STATIC_MUTEX_INIT;
@@ -41,7 +41,7 @@ static int ref_familyMonospace = 0;
 
 /* Family and collections font functions */
 
-void
+static void
 gdip_createFontFamily (GpFontFamily **family)
 {
 	GpFontFamily *result = (GpFontFamily *) GdipAlloc (sizeof (GpFontFamily));
@@ -140,12 +140,12 @@ GdipDeletePrivateFontCollection (GpFontCollection **font_collection)
 GpStatus
 GdipPrivateAddFontFile (GpFontCollection *font_collection,  GDIPCONST WCHAR *filename)
 {
-	unsigned char *file;
+	BYTE *file;
 	
 	if (!font_collection || !filename)
 		return InvalidParameter;
     
-	file = (unsigned char*) ucs2_to_utf8 ((const gunichar2 *)filename, -1);
+	file = (BYTE*) ucs2_to_utf8 ((const gunichar2 *)filename, -1);
 	if (!file)
 		return OutOfMemory;
 
@@ -159,7 +159,7 @@ GdipPrivateAddFontFile (GpFontCollection *font_collection,  GDIPCONST WCHAR *fil
 GpStatus 
 GdipDeleteFontFamily (GpFontFamily *fontFamily)
 {
-	bool delete = TRUE;
+	BOOL delete = TRUE;
 	
 	if (!fontFamily)
 		return InvalidParameter;
@@ -203,7 +203,7 @@ GdipDeleteFontFamily (GpFontFamily *fontFamily)
 	return Ok;
 }
 
-void
+static void
 gdip_createPrivateFontSet (GpFontCollection *font_collection)
 {
 	FcObjectSet *os = FcObjectSetBuild (FC_FAMILY, FC_FOUNDRY, NULL);
@@ -287,7 +287,7 @@ create_pattern_from_name (char* name)
 		
 	/* find the family we want */
 	val.type = FcTypeString;
-	val.u.s = (unsigned char*)name;
+	val.u.s = (BYTE*)name;
 	if (!FcPatternAdd (name_pattern, FC_FAMILY, val, TRUE)) {
 		FcPatternDestroy (name_pattern);
 		return NULL;
@@ -359,10 +359,10 @@ create_fontfamily_from_name (char* name, GpFontFamily **fontFamily)
 	return status;
 }
 
-static
-gboolean free_cached_pattern (gpointer key, gpointer value, gpointer user)
+static BOOL
+free_cached_pattern (gpointer key, gpointer value, gpointer user)
 {
-	g_free (value);
+	g_free (key);
 	FcPatternDestroy ((FcPattern*) value);
 	return TRUE;
 }
@@ -497,13 +497,13 @@ GdipGetGenericFontFamilySerif (GpFontFamily **nativeFamily)
 GpStatus
 GdipGetGenericFontFamilyMonospace (GpFontFamily **nativeFamily)
 {
-	const WCHAR Serif[] = {'S','e','r','i','f', 0};
+	const WCHAR Monospace[] = {'C','o','u','r','i', 'e', 'r', ' ', 'N', 'e', 'w', 0};
 	GpStatus status = Ok;
 	
 	g_static_mutex_lock (&generic);
 
 	if (ref_familyMonospace == 0)
-		status = GdipCreateFontFamilyFromName (Serif, NULL, &familyMonospace);
+		status = GdipCreateFontFamilyFromName (Monospace, NULL, &familyMonospace);
 
 	if (status == Ok)
 		ref_familyMonospace++;
@@ -516,7 +516,7 @@ GdipGetGenericFontFamilyMonospace (GpFontFamily **nativeFamily)
 	return status;
 }
 
-FT_Face
+static FT_Face
 gdip_cairo_ft_font_lock_face (cairo_font_face_t *cairofnt, cairo_scaled_font_t **scaled_ft)
 {
 	FT_Face face;
@@ -547,164 +547,111 @@ gdip_cairo_ft_font_lock_face (cairo_font_face_t *cairofnt, cairo_scaled_font_t *
 	return face;
 }
 
-void
+static void
 gdip_cairo_ft_font_unlock_face (cairo_scaled_font_t* scaled_ft)
 {
 	cairo_ft_scaled_font_unlock_face (scaled_ft);
 	cairo_scaled_font_destroy (scaled_ft);
 }
 
-GpStatus
-GdipGetEmHeight (GDIPCONST GpFontFamily *family, GpFontStyle style, short *EmHeight)
+static GpStatus
+gdip_get_fontfamily_details (GpFontFamily *family, FontStyle style)
 {
-	short rslt = 0;
 	GpFont *font = NULL;
+	GpStatus status = GdipCreateFont (family, 0.0f, style, UnitPoint, &font);
+
+	if ((status == Ok) && font) {
+		cairo_scaled_font_t* scaled_ft;
+
+		FT_Face face = gdip_cairo_ft_font_lock_face (font->cairofnt, &scaled_ft);
+		if (face) {
+			TT_VertHeader *pVert = FT_Get_Sfnt_Table (face, ft_sfnt_vhea);
+			TT_HoriHeader *pHori = FT_Get_Sfnt_Table (face, ft_sfnt_hhea);
+
+			if (pVert) {
+				family->height = pVert->yMax_Extent;
+			} else {
+				family->height = face->units_per_EM;
+			}
+
+			if (pHori) {
+				/* FIXME: FT docs seems to suggest the use of the OS/2 table */
+				family->cellascent = pHori->Ascender;
+				family->celldescent = -pHori->Descender;
+				family->linespacing = pHori->Ascender - pHori->Descender + pHori->Line_Gap;
+			} else {
+				family->cellascent = face->ascender;
+				family->celldescent = face->descender;
+				family->linespacing = face->height;
+			}
+
+			gdip_cairo_ft_font_unlock_face (scaled_ft);
+		} else
+			status = FontFamilyNotFound;
+	}
+	if (font)
+		GdipDeleteFont (font);
+	return status;
+}
+
+GpStatus
+GdipGetEmHeight (GDIPCONST GpFontFamily *family, int style, guint16 *EmHeight)
+{
+	GpStatus status = Ok;
 
 	if (!family || !EmHeight)
 		return InvalidParameter;
 
-	if (family->height != -1) {
-		*EmHeight = family->height;
-		return Ok;
-	}
+	if (family->height == -1)
+		status = gdip_get_fontfamily_details ((GpFontFamily*)family, style);
 
-	GdipCreateFont (family, 0.0f, style, UnitPoint, &font);
-
-	if (font) {
-		cairo_scaled_font_t* scaled_ft;
-		FT_Face	face;
-		TT_VertHeader *pVert;
-
-		face = gdip_cairo_ft_font_lock_face (font->cairofnt, &scaled_ft);
-		if (face) {
-			pVert = FT_Get_Sfnt_Table(face, ft_sfnt_vhea);
-			if (pVert) {
-				rslt = pVert->yMax_Extent;
-			} else if (face) {
-				rslt = face->units_per_EM;
-			}
-			gdip_cairo_ft_font_unlock_face (scaled_ft);
-		}
-		GdipDeleteFont (font);
-	}
-
-	*EmHeight = rslt;
-	((GpFontFamily *)family)->height = rslt;
-	return Ok;
+	*EmHeight = family->height;
+	return status;
 }
 
 GpStatus
-GdipGetCellAscent (GDIPCONST GpFontFamily *family, GpFontStyle style, short *CellAscent)
+GdipGetCellAscent (GDIPCONST GpFontFamily *family, int style, guint16 *CellAscent)
 {
-	short rslt = 0;
-	GpFont *font = NULL;
+	GpStatus status = Ok;
 
 	if (!family || !CellAscent)
 		return InvalidParameter;
 
-	if (family->cellascent != -1) {
-		*CellAscent = family->cellascent;
-		return Ok;
-	}
+	if (family->cellascent == -1)
+		status = gdip_get_fontfamily_details ((GpFontFamily*)family, style);
 
-	GdipCreateFont (family, 0.0f, style, UnitPoint, &font);
-
-	if (font){
-		cairo_scaled_font_t* scaled_ft;
-                FT_Face face;
-		TT_HoriHeader *pHori;
-
-		face = gdip_cairo_ft_font_lock_face (font->cairofnt, &scaled_ft);
-		if (face) {
-			pHori = FT_Get_Sfnt_Table (face, ft_sfnt_hhea);
-			if (pHori)
-				rslt = pHori->Ascender;
-
-			gdip_cairo_ft_font_unlock_face (scaled_ft);
-		}
-		GdipDeleteFont (font);
-	}
-
-	*CellAscent = rslt;
-	((GpFontFamily *)family)->cellascent = rslt;
-	return Ok;         
+	*CellAscent = family->cellascent;
+	return status;
 }
 
 GpStatus
-GdipGetCellDescent (GDIPCONST GpFontFamily *family, GpFontStyle style, short *CellDescent)
+GdipGetCellDescent (GDIPCONST GpFontFamily *family, int style, guint16 *CellDescent)
 {
-	short rslt = 0;
-	GpFont *font = NULL;
+	GpStatus status = Ok;
 
 	if (!family || !CellDescent)
 		return InvalidParameter;
 
-	if (family->celldescent != -1) {
-		*CellDescent = family->celldescent;
-		return Ok;
-	}
+	if (family->celldescent == -1)
+		status = gdip_get_fontfamily_details ((GpFontFamily*)family, style);
 
-	GdipCreateFont (family, 0.0f, style, UnitPoint, &font);
-
-	if (font){
-		cairo_scaled_font_t* scaled_ft;
-                FT_Face face;
-		TT_HoriHeader* pHori;
-
-		face = gdip_cairo_ft_font_lock_face (font->cairofnt, &scaled_ft);
-		if (face) {
-			pHori = FT_Get_Sfnt_Table (face, ft_sfnt_hhea);
-			if (pHori)
-				rslt = -pHori->Descender;
-
-			gdip_cairo_ft_font_unlock_face (scaled_ft);
-		}
-		GdipDeleteFont (font);
-	}
-
-	*CellDescent = rslt;
-	((GpFontFamily *)family)->celldescent = rslt;
-	return Ok;         
+	*CellDescent = family->celldescent;
+	return status;
 }
 
 GpStatus
-GdipGetLineSpacing (GDIPCONST GpFontFamily *family, GpFontStyle style, short *LineSpacing)
+GdipGetLineSpacing (GDIPCONST GpFontFamily *family, int style, guint16 *LineSpacing)
 {
-	short rslt = 0;
-	GpFont *font = NULL;
+	GpStatus status = Ok;
 
 	if (!family || !LineSpacing)
 		return InvalidParameter;
 
-	if (family->linespacing != -1) {
-		*LineSpacing = family->linespacing;
-		return Ok;
-	}
+	if (family->linespacing == -1)
+		status = gdip_get_fontfamily_details ((GpFontFamily*)family, style);
 
-	GdipCreateFont (family, 0.0f, style, UnitPoint, &font);
-
-	if (font){
-		cairo_scaled_font_t* scaled_ft;
-                FT_Face face;
-		TT_HoriHeader *pHori;
-
-		face = gdip_cairo_ft_font_lock_face (font->cairofnt, &scaled_ft);
-		if (face) {
-			pHori = FT_Get_Sfnt_Table (face, ft_sfnt_hhea);
-			if (pHori) {
-				rslt = pHori->Ascender + (-pHori->Descender) + pHori->Line_Gap;
-			} else if (face) {
-				rslt = face->height;
-			}
-
-			gdip_cairo_ft_font_unlock_face(scaled_ft);
-		}
-		GdipDeleteFont (font);
-	}
-
-	*LineSpacing = rslt;
-	((GpFontFamily *)family)->linespacing = rslt;
-	return Ok;
+	*LineSpacing = family->linespacing;
+	return status;
 }
 
 GpStatus
@@ -718,7 +665,7 @@ GdipIsStyleAvailable (GDIPCONST GpFontFamily *family, int style, BOOL *IsStyleAv
 }
 
 /* Font functions */
-cairo_font_face_t *
+static cairo_font_face_t*
 gdip_face_create (const char *family, 
 			cairo_font_slant_t   slant, 
 			cairo_font_weight_t  weight,
@@ -727,18 +674,17 @@ gdip_face_create (const char *family,
 	cairo_surface_t *surface;
 	cairo_font_face_t *face;
 
-	surface = cairo_image_surface_create_for_data ((unsigned char *)NULL, CAIRO_FORMAT_ARGB32, 0, 0, 0);
+	surface = cairo_image_surface_create_for_data ((BYTE*)NULL, CAIRO_FORMAT_ARGB32, 0, 0, 0);
 	*ct = cairo_create (surface);
 	cairo_select_font_face (*ct, (const char *) family, slant, weight);
 	face = cairo_get_font_face (*ct);
 	cairo_surface_destroy (surface);
 	return face;
-
 }
 
 // coverity[+alloc : arg-*4]
 GpStatus
-GdipCreateFont (GDIPCONST GpFontFamily* family, float emSize, GpFontStyle style, Unit unit,  GpFont **font)
+GdipCreateFont (GDIPCONST GpFontFamily* family, float emSize, int style, Unit unit, GpFont **font)
 {
 	GpStatus status;
 	FcChar8* str;
@@ -797,17 +743,6 @@ GdipCreateFont (GDIPCONST GpFontFamily* family, float emSize, GpFontStyle style,
 	return Ok;
 }
 
-GpStatus
-GdipGetHfont(GpFont* font, void **Hfont)
-{
-	if (font) {
-		*Hfont=font;
-		return(Ok);
-	}
-	return InvalidParameter;
-}
-
-
 GpStatus
 GdipDeleteFont (GpFont* font)
 {
@@ -816,8 +751,8 @@ GdipDeleteFont (GpFont* font)
 
 	cairo_destroy (font->ct);
 
-	GdipFree ((void *)font->face);
-	GdipFree ((void *)font);
+	GdipFree (font->face);
+	GdipFree (font);
 	return Ok;	       
 }
 
@@ -828,7 +763,7 @@ GdipCreateFontFromDC(void *hdc, GpFont **font)
 }
 
 static GpStatus
-gdip_logfont_from_font(GpFont *font, GpGraphics *graphics, void *lf, bool ucs2)
+gdip_logfont_from_font (GpFont *font, GpGraphics *graphics, void *lf, BOOL ucs2)
 {
 	LOGFONTA		*logFont;
 
@@ -963,19 +898,19 @@ GdipCreateFontFromHfontA(void *hfont, GpFont **font, void *lf)
 }
 
 GpStatus
-GdipGetLogFontW(GpFont *font, GpGraphics *graphics, void *lf)
+GdipGetLogFontW (GpFont *font, GpGraphics *graphics, LOGFONTW *logfontW)
 {
-	return gdip_logfont_from_font (font, graphics, lf, TRUE);
+	return gdip_logfont_from_font (font, graphics, logfontW, TRUE);
 }
 
 GpStatus
-GdipGetLogFontA(GpFont *font, GpGraphics *graphics, void *lf)
+GdipGetLogFontA (GpFont *font, GpGraphics *graphics, LOGFONTA *logfontA)
 {
-	return gdip_logfont_from_font (font, graphics, lf, FALSE);
+	return gdip_logfont_from_font (font, graphics, logfontA, FALSE);
 }
 
-GpStatus
-gdip_create_font_from_logfont(void *hdc, void *lf, GpFont **font, bool ucs2)
+static GpStatus
+gdip_create_font_from_logfont (void *hdc, void *lf, GpFont **font, BOOL ucs2)
 {
 	GpFont			*result;
 	cairo_font_face_t	*face;
@@ -1019,7 +954,7 @@ gdip_create_font_from_logfont(void *hdc, void *lf, GpFont **font, bool ucs2)
 	}
 
 	if (ucs2) {
-		result->face = (unsigned char*) ucs2_to_utf8 ((const gunichar2 *)logfont->lfFaceName, -1);
+		result->face = (BYTE*) ucs2_to_utf8 ((const gunichar2 *)logfont->lfFaceName, -1);
 		if (!result->face){
 			GdipFree (result);
 			return OutOfMemory;
@@ -1091,60 +1026,52 @@ GdipPrivateAddMemoryFont(GpFontCollection *fontCollection, GDIPCONST void *memor
 	return result ? Ok : FileNotFound;
 }
 
-
-bool
-GetFontMetrics(GpGraphics *graphics, GpFont *font, int *ascent, int *descent)
-{
-	cairo_font_extents_t	font_extent;
-
-	cairo_set_font_face (graphics->ct, (cairo_font_face_t*) font->cairofnt);
-	cairo_set_font_size (graphics->ct, font->sizeInPixels);
-	cairo_font_extents (graphics->ct, &font_extent);
-
-	if (ascent) {
-		*ascent = (int)font_extent.ascent;
-	}
-
-	if (descent) {
-		*descent = (int)font_extent.descent;
-	}
-
-	return TRUE;
-}
-
-
 GpStatus
 GdipGetFontHeight (GDIPCONST GpFont *font, GDIPCONST GpGraphics *graphics, float *height)
 {
-	short emHeight, lineSpacing;
-	float emSize;
+	GpStatus status;
+	guint16 emHeight, lineSpacing;
+	float emSize, h;
 
 	if (!font || !height || !graphics)
 		return InvalidParameter;
 
+	status = GdipGetEmHeight (font->family, font->style, &emHeight);
+	if (status != Ok)
+		return status;
+
+	status = GdipGetLineSpacing (font->family, font->style, &lineSpacing);
+	if (status != Ok)
+		return status;
+
 	/* Operations in display dpi's */	
 	emSize = gdip_unit_conversion (font->unit, UnitPixel, gdip_get_display_dpi (), gtMemoryBitmap, font->emSize);
-	GdipGetEmHeight (font->family, font->style, &emHeight);
-	GdipGetLineSpacing (font->family, font->style, &lineSpacing);
 
-	*height = lineSpacing * (emSize / emHeight);
-	*height = gdip_unit_conversion (UnitPixel, graphics->page_unit, gdip_get_display_dpi (), graphics->type, *height);
+	h = lineSpacing * (emSize / emHeight);
+	*height = gdip_unit_conversion (UnitPixel, graphics->page_unit, gdip_get_display_dpi (), graphics->type, h);
 	return Ok;
 }
 
 GpStatus
 GdipGetFontHeightGivenDPI (GDIPCONST GpFont *font, float dpi, float *height)
 {
-	short emHeight, lineSpacing;
+	GpStatus status;
+	guint16 emHeight, lineSpacing;
+	float h;
 
 	if (!font || !height)
 		return InvalidParameter;
 
-	GdipGetEmHeight (font->family, font->style, &emHeight);
-	GdipGetLineSpacing (font->family, font->style, &lineSpacing);
-	*height = lineSpacing * (font->emSize / emHeight);
-	*height = gdip_unit_conversion (font->unit, UnitInch, dpi, gtMemoryBitmap, *height);
-	*height = *height * dpi;
+	status = GdipGetEmHeight (font->family, font->style, &emHeight);
+	if (status != Ok)
+		return status;
+
+	status = GdipGetLineSpacing (font->family, font->style, &lineSpacing);
+	if (status != Ok)
+		return status;
+
+	h = lineSpacing * (font->emSize / emHeight);
+	*height = gdip_unit_conversion (font->unit, UnitInch, dpi, gtMemoryBitmap, h) * dpi;
 	return Ok;
 }
 
