@@ -25,105 +25,62 @@
  *   Sebastien Pouliot  <sebastien@ximian.com>
  */
 
-#include "general.h"
-#include <cairo.h>
-#include <math.h>
-
-/* #define DEBUG_MEMLEAKS	1 */
-
-/* Startup / shutdown */
-
-struct startupInput
-{
-	unsigned int version;             
-	void       * ptr; 
-	int          threadOpt;
-	int          codecOpt;
-};
-
-
-struct startupOutput
-{
-	void *hook;
-	void *unhook;
-};
+#include "general-private.h"
+#include "codecs-private.h"
+#include "graphics-private.h"
+#include "font-private.h"
 
 /* large table to avoid a division and three multiplications when premultiplying alpha into R, G and B */
 #include "alpha-premul-table.inc"
 
-static GList* g_mem_allocations;
 static BOOL startup = FALSE;
 
-extern void cairo_test_xlib_disable_render();
-
 GpStatus 
-GdiplusStartup(unsigned long *token, const struct startupInput *input, struct startupOutput *output)
+GdiplusStartup (ULONG_PTR *token, const GdiplusStartupInput *input, GdiplusStartupOutput *output)
 {
 	GpStatus status = Ok;
 	/* don't initialize multiple time, e.g. for each appdomain */
 	if (!startup) {
 		startup = TRUE;
-		g_mem_allocations = NULL;
 		status = initCodecList ();
 		if (status != Ok)
 			return status;
 		FcInit ();
 		*token = 1;
 		gdip_get_display_dpi();
-//		cairo_test_xlib_disable_render();
 	}
 	return status;
 }
 
 void 
-GdiplusShutdown(unsigned long *token)
+GdiplusShutdown (ULONG_PTR token)
 {
-#ifdef DEBUG_MEMLEAKS
-	GList* list = NULL;
-#endif
 	if (startup) {
 		releaseCodecList ();
 		gdip_font_clear_pattern_cache ();
+#if HAVE_FCFINI
+		FcFini ();
+#endif
 		startup = FALSE; /* in case we want to restart it */
 	}
-#ifdef DEBUG_MEMLEAKS
-	for (list = g_list_first (g_mem_allocations); list != NULL; list = g_list_next (list)) {
-		printf ("Memory block not free'd at %x\n", list->data);
-	}
-#endif
 }
 
 
 /* Memory */
 void *
-GdipAlloc (int size)
+GdipAlloc (size_t size)
 {
-#ifdef DEBUG_MEMLEAKS		
-	void* block = malloc (size);
-	g_mem_allocations =  g_list_append (g_mem_allocations, block);
-	printf ("alloc %x %u\n", block, size);
-	return block;
-#else
 	return malloc (size);
-#endif
 }
 
 void *
-GdipCalloc (size_t nelem, size_t elsize)
+gdip_calloc (size_t nelem, size_t elsize)
 {
-#ifdef DEBUG_MEMLEAKS		
-	void* block = calloc (nelem, elsize);
-	g_mem_allocations =  g_list_append (g_mem_allocations, block);
-	printf ("calloc %x %u\n", block, nelem);
-	return block;
-#else
 	return calloc (nelem, elsize);
-#endif
-
 }
 
 void *
-GdipRealloc (void *org, int size)
+gdip_realloc (void *org, int size)
 {
 	return realloc (org, size);
 }
@@ -131,20 +88,7 @@ GdipRealloc (void *org, int size)
 void 
 GdipFree (void *ptr)
 {
-#ifdef DEBUG_MEMLEAKS	
-	GList* list = NULL;
-
-	list = g_list_find (g_list_first (g_mem_allocations), ptr);
-	if (list != NULL) {
-		list  = g_list_delete_link  (g_list_first (g_mem_allocations), list);
-		g_mem_allocations = g_list_last (list);
-	}
-
-	printf ("free %x\n", ptr);
 	free (ptr);
-#else
-	free (ptr);
-#endif
 }
 
 /* Helpers */
@@ -174,47 +118,8 @@ gdip_get_status (cairo_status_t status)
         }
 }
 
-/* Converts the given interpolation value to cairo_filter_t */
-cairo_filter_t
-gdip_get_cairo_filter (InterpolationMode imode)
-{
-	cairo_filter_t filter;
-
-	switch (imode) {
-
-	case InterpolationModeHighQuality:
-	case InterpolationModeHighQualityBilinear:
-	case InterpolationModeHighQualityBicubic:
-		filter = CAIRO_FILTER_BEST;
-		break;
-
-	case InterpolationModeNearestNeighbor:
-		filter = CAIRO_FILTER_NEAREST;
-		break;
-
-	case InterpolationModeBilinear:
-		filter = CAIRO_FILTER_BILINEAR;
-		break;
-
-	case InterpolationModeBicubic:
-		filter = CAIRO_FILTER_GAUSSIAN;
-		break;
-		
-	case InterpolationModeLowQuality:
-		filter = CAIRO_FILTER_FAST;
-		break;
-
-	case InterpolationModeDefault:
-	default:
-		filter = CAIRO_FILTER_GOOD;
-		break;
-	}
-
-	return filter;
-}
-
 float
-gdip_get_display_dpi()
+gdip_get_display_dpi ()
 {
 	static float dpis = 0;
 	Display* display;
@@ -331,33 +236,8 @@ gdip_closed_curve_tangents (int terms, const GpPointF *points, int count, float
         return tangents;
 }
 
-void
-gdip_rect_expand_by (GpRectF *rect, GpPointF *point)
-{
-    /* This method is somewhat stupid, because GpRect is x,y width,height,
-     * instead of x0,y0 x1,y1.
-     */
-    float x0 = rect->X;
-    float y0 = rect->Y;
-    float x1 = x0 + rect->Width;
-    float y1 = y0 + rect->Height;
-
-    if (point->X < x0)
-        x0 = point->X;
-    else if (point->X > x1)
-        x1 = point->X;
-
-    if (point->Y < y0)
-        y0 = point->Y;
-    else if (point->Y > y1)
-        y1 = point->Y;
-
-    rect->X = x0;
-    rect->Y = y0;
-    rect->Width = (x1 - x0);
-    rect->Height = (y1 - y0);
-}
-
+/* re-enabled if/when required */
+#if FALSE
 /* this function comes from fcmp.sf.net */
 int
 fcmp (double x1, double x2, double epsilon)
@@ -401,11 +281,21 @@ fcmp (double x1, double x2, double epsilon)
         else /* -delta <= difference <= delta */
                 return 0;  /* x1 == x2 */
 }
+#endif
+
+/* note: round[f] is C99 */
+int
+iround (float d)
+{
+	double int_part = floor (d);
+	return ((d - int_part) >= 0.5) ? int_part + 1.0 : int_part;
+}
 
 /* This function is used by gradient brushes for calculating the erf required for 
  * calculating the integral of the normal distribution equation.
  */
-float gdip_erf (float x, float std, float mean)
+float
+gdip_erf (float x, float std, float mean)
 {
 	/* std refers to standard deviation
 	 *
@@ -478,7 +368,7 @@ ucs2_to_utf8(const gunichar2 *ucs2, int length) {
 	return utf8;
 }
 
-bool
+BOOL
 utf8_to_ucs2(const gchar *utf8, gunichar2 *ucs2, int ucs2_len) {
 	int 		i;
 	glong		items_read;
@@ -512,27 +402,29 @@ utf8_to_ucs2(const gchar *utf8, gunichar2 *ucs2, int ucs2_len) {
 }
 
 int
-utf8_encode_ucs2char(gunichar2 unichar, unsigned char *dest)
+utf8_encode_ucs2char(gunichar2 unichar, BYTE *dest)
 {
 	if (unichar < 0x0080) {					/* 0000-007F */
-		dest[0] = (unsigned char)(unichar);
+		dest[0] = (BYTE)(unichar);
 		return (1);
 	}
 	if(unichar < 0x0800) {					/* 0080-07FF */
-		dest[0] = (unsigned char)(0xC0 | ((unichar & 0x07C0) >> 6));
-		dest[1] = (unsigned char)(0x80 | (unichar & 0x003F));
+		dest[0] = (BYTE)(0xC0 | ((unichar & 0x07C0) >> 6));
+		dest[1] = (BYTE)(0x80 | (unichar & 0x003F));
 		return (2);
 	}
 								/* 0800-FFFF */
-	dest[0] = (unsigned char)(0xE0 | ((unichar & 0xF000) >> 12));
-	dest[1] = (unsigned char)(0x80 | ((unichar & 0x0FC0) >> 6));
-	dest[2] = (unsigned char)(0x80 | (unichar & 0x003F));
+	dest[0] = (BYTE)(0xE0 | ((unichar & 0xF000) >> 12));
+	dest[1] = (BYTE)(0x80 | ((unichar & 0x0FC0) >> 6));
+	dest[2] = (BYTE)(0x80 | (unichar & 0x003F));
 	return (3);	
 }
 
+/* re-enabled if/when required */
+#if FALSE
 /* This function only handles UCS-2 */
 int
-utf8_decode_ucs2char(const unsigned char *src, gunichar2 *uchar)
+utf8_decode_ucs2char (const BYTE *src, gunichar2 *uchar)
 {
 	if (src[0] <= 0x7F) {			/* 0000-007F: one byte (0xxxxxxx) */
 		*uchar = (gunichar2)src[0];
@@ -549,20 +441,7 @@ utf8_decode_ucs2char(const unsigned char *src, gunichar2 *uchar)
 		((((gunichar2)src[2]) & 0x003F) << 0);
 	return (3);
 }
-
-cairo_content_t
-from_cairoformat_to_content (cairo_format_t format)
-{
-    	switch (format) {
-    	case CAIRO_FORMAT_RGB24:
-		return CAIRO_CONTENT_COLOR;
-	case CAIRO_FORMAT_A8:
-		return CAIRO_CONTENT_ALPHA;
-	case CAIRO_FORMAT_ARGB32:
-	default:
-		return CAIRO_CONTENT_COLOR_ALPHA;
-    }
-}
+#endif
 
 GpStatus
 gdip_get_pattern_status (cairo_pattern_t *pat)
