@@ -1,7 +1,7 @@
 /*
  * lineargradientbrush.c
  *
- * Copyright (C) Novell, Inc. 2003-2004.
+ * Copyright (C) 2003-2004,2007 Novell, Inc. http://www.novell.com
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy of this software 
  * and associated documentation files (the "Software"), to deal in the Software without restriction, 
@@ -23,11 +23,9 @@
  *
  */
 
-#include <math.h>
-
-#include "gdip.h"
-#include "matrix.h"
-#include "lineargradientbrush.h"
+#include "lineargradientbrush-private.h"
+#include "graphics-private.h"
+#include "matrix-private.h"
 
 static GpStatus gdip_linear_gradient_setup (GpGraphics *graphics, GpBrush *brush);
 static GpStatus gdip_linear_gradient_clone_brush (GpBrush *brush, GpBrush **clonedBrush);
@@ -43,7 +41,7 @@ static BrushClass vtable = { BrushTypeLinearGradient,
 			     gdip_linear_gradient_clone_brush,
 			     gdip_linear_gradient_destroy };
 
-void
+static void
 gdip_linear_gradient_init (GpLineGradient *linear)
 {
 	gdip_brush_init (&linear->base, &vtable);
@@ -64,7 +62,7 @@ gdip_linear_gradient_init (GpLineGradient *linear)
 	linear->pattern = NULL;
 }
 
-GpLineGradient*
+static GpLineGradient*
 gdip_linear_gradient_new (void)
 {
 	GpLineGradient *result = (GpLineGradient *) GdipAlloc (sizeof (GpLineGradient));
@@ -171,12 +169,10 @@ gdip_linear_gradient_clone_brush (GpBrush *brush, GpBrush **clonedBrush)
 GpStatus
 gdip_linear_gradient_destroy (GpBrush *brush)
 {
-	GpLineGradient *linear;
+	/* a. the NULL check for brush is done by the caller, GdipDeleteBrush */
+	/* b. brush itself is freed by the caller */
 
-	if (!brush)
-		return InvalidParameter;
-
-	linear = (GpLineGradient *) brush;
+	GpLineGradient *linear = (GpLineGradient *) brush;
 
 	if (linear->blend) {
 		if (linear->blend->count > 0) {
@@ -201,12 +197,10 @@ gdip_linear_gradient_destroy (GpBrush *brush)
 		linear->pattern = NULL;
 	}
 
-	GdipFree (linear);
-
 	return Ok;
 }
 
-void
+static void
 add_color_stops_from_blend (cairo_pattern_t *pattern, Blend *blend, ARGB *colors)
 {
 	int index;
@@ -236,7 +230,7 @@ add_color_stops_from_blend (cairo_pattern_t *pattern, Blend *blend, ARGB *colors
 	}
 }
 
-void
+static void
 add_color_stops_from_interpolation_colors (cairo_pattern_t *pattern, InterpolationColors *presetColors)
 {
 	int index;
@@ -260,7 +254,7 @@ add_color_stops_from_interpolation_colors (cairo_pattern_t *pattern, Interpolati
 	}
 }
 
-void
+static void
 add_color_stops (cairo_pattern_t *pattern, ARGB *colors)
 {
 	double r, g, b, a;
@@ -282,7 +276,7 @@ add_color_stops (cairo_pattern_t *pattern, ARGB *colors)
 	cairo_pattern_add_color_stop_rgba (pattern, 1.0, r / 255, g / 255, b / 255, a / 255);
 }
 
-GpStatus
+static GpStatus
 create_tile_linear (GpGraphics *graphics, cairo_t *ct, GpLineGradient *linear)
 {
 	GpStatus status;
@@ -997,9 +991,6 @@ GdipSetLineLinearBlend (GpLineGradient *brush, float focus, float scale)
 	return Ok;
 }
 
-/* This is defined in general.c */
-float gdip_erf (float x, float std, float mean);
-
 GpStatus
 GdipSetLineSigmaBlend (GpLineGradient *brush, float focus, float scale)
 {
