@@ -28,17 +28,17 @@
 #if HAVE_CONFIG_H
 #include <config.h>
 #endif
+#include "win32structs.h"
+
+GUID gdip_png_image_format_guid = {0xb96b3cafU, 0x0728U, 0x11d3U, {0x9d, 0x7b, 0x00, 0x00, 0xf8, 0x1e, 0xf3, 0x2e}};
 
 #ifdef HAVE_LIBPNG
 
 #include <png.h>
-#include "general.h"
+#include "gdiplus-private.h"
 #include "pngcodec.h"
-
-
 #include <setjmp.h>
 
-
 /* Codecinfo related data*/
 static ImageCodecInfo png_codec;
 static const WCHAR png_codecname[] = {'B', 'u', 'i','l', 't', '-','i', 'n', ' ', 'P', 'N', 'G', 0}; /* Built-in PNG */
@@ -59,7 +59,7 @@ gdip_getcodecinfo_png ()
         png_codec.FormatDescription = (const WCHAR*) png_format;
         png_codec.FilenameExtension = (const WCHAR*) png_extension;
         png_codec.MimeType = (const WCHAR*) png_mimetype;
-        png_codec.Flags = Encoder | Decoder | SupportBitmap | Builtin;
+        png_codec.Flags = ImageCodecFlagsEncoder | ImageCodecFlagsDecoder | ImageCodecFlagsSupportBitmap | ImageCodecFlagsBuiltin;
         png_codec.Version = 1;
         png_codec.SigCount = 1;
         png_codec.SigSize = 8;
@@ -125,8 +125,8 @@ gdip_load_png_properties (png_structp png_ptr, png_infop info_ptr, png_infop end
 		int		compression_type;
 
 		if (png_get_iCCP(png_ptr, info_ptr, &name, &compression_type, &profile, &proflen)) {
-			gdip_bitmapdata_property_add_ASCII(bitmap_data, ICCProfileDescriptor, (unsigned char *)name);
-			gdip_bitmapdata_property_add_byte(bitmap_data, ICCProfile, (byte)compression_type);
+			gdip_bitmapdata_property_add_ASCII(bitmap_data, PropertyTagICCProfileDescriptor, (BYTE*)name);
+			gdip_bitmapdata_property_add_byte(bitmap_data, PropertyTagICCProfile, (BYTE)compression_type);
 		}
 	}
 #endif
@@ -136,7 +136,7 @@ gdip_load_png_properties (png_structp png_ptr, png_infop info_ptr, png_infop end
 		double	gamma;
 
 		if (png_get_gAMA(png_ptr, info_ptr, &gamma)) {
-			gdip_bitmapdata_property_add_rational(bitmap_data, Gamma, 100000, gamma * 100000);
+			gdip_bitmapdata_property_add_rational(bitmap_data, PropertyTagGamma, 100000, gamma * 100000);
 		}
 	}
 #endif
@@ -153,7 +153,7 @@ gdip_load_png_properties (png_structp png_ptr, png_infop info_ptr, png_infop end
 		double	blue_y;
 
 		if (png_get_cHRM(png_ptr, info_ptr, &white_x, &white_y, &red_x, &red_y, &green_x, &green_y, &blue_x, &blue_y)) {
-			unsigned char	*buffer;
+			BYTE *buffer;
 			guint32		*ptr;
 
 			buffer = GdipAlloc(6 * (sizeof(png_uint_32) + sizeof(png_uint_32)));
@@ -175,14 +175,15 @@ gdip_load_png_properties (png_structp png_ptr, png_infop info_ptr, png_infop end
 				ptr[10] = (guint32)(blue_y * 100000);
 				ptr[11] = 100000;
 
-				gdip_bitmapdata_property_add(bitmap_data, PrimaryChromaticities, 6 * (sizeof(guint32) + sizeof(guint32)), TypeRational, buffer);
+				gdip_bitmapdata_property_add (bitmap_data, PropertyTagPrimaryChromaticities, 
+					6 * (sizeof(guint32) + sizeof(guint32)), PropertyTagTypeRational, buffer);
 
 				ptr[0] = (guint32)(white_x * 100000);
 				ptr[1] = 1000000;
 				ptr[2] = (guint32)(white_y * 100000);
 				ptr[3] = 100000;
-				gdip_bitmapdata_property_add(bitmap_data, WhitePoint, 2 * (sizeof(guint32) + sizeof(guint32)), TypeRational, buffer);
-
+				gdip_bitmapdata_property_add (bitmap_data, PropertyTagWhitePoint, 
+					2 * (sizeof(guint32) + sizeof(guint32)), PropertyTagTypeRational, buffer);
 
 				GdipFree(buffer);
 			}
@@ -197,9 +198,9 @@ gdip_load_png_properties (png_structp png_ptr, png_infop info_ptr, png_infop end
 		png_uint_32	res_y;
 
 		if (png_get_pHYs(png_ptr, info_ptr, &res_x, &res_y, &unit_type)) {
-			gdip_bitmapdata_property_add_byte(bitmap_data, PixelUnit, (byte)unit_type);
-			gdip_bitmapdata_property_add_long(bitmap_data, PixelPerUnitX, res_x);
-			gdip_bitmapdata_property_add_long(bitmap_data, PixelPerUnitY, res_y);
+			gdip_bitmapdata_property_add_byte(bitmap_data, PropertyTagPixelUnit, (BYTE)unit_type);
+			gdip_bitmapdata_property_add_long(bitmap_data, PropertyTagPixelPerUnitX, res_x);
+			gdip_bitmapdata_property_add_long(bitmap_data, PropertyTagPixelPerUnitY, res_y);
 		}
 	}
 #endif
@@ -211,7 +212,7 @@ gdip_load_png_properties (png_structp png_ptr, png_infop info_ptr, png_infop end
 
 		if (png_get_text(png_ptr, info_ptr, &text_ptr, &num_text)) {
 			if (num_text > 0) {
-				gdip_bitmapdata_property_add_ASCII(bitmap_data, ExifUserComment, (unsigned char *)text_ptr[0].text);
+				gdip_bitmapdata_property_add_ASCII(bitmap_data, PropertyTagExifUserComment, (BYTE*)text_ptr[0].text);
 			}
 		}
 	}
@@ -226,9 +227,12 @@ gdip_load_png_image_from_file_or_stream (FILE *fp, GetBytesDelegate getBytesFunc
 	png_structp	png_ptr = NULL;
 	png_infop	info_ptr = NULL;
 	png_infop	end_info_ptr = NULL;
-	guchar		*rawdata = NULL;
+	BYTE		*rawdata = NULL;
 	GpImage		*result = NULL;
 	GpStatus	status = InvalidParameter;
+	int		bit_depth;
+	int		channels;
+	BYTE		color_type;
 
 	png_ptr = png_create_read_struct (PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
 
@@ -236,7 +240,6 @@ gdip_load_png_image_from_file_or_stream (FILE *fp, GetBytesDelegate getBytesFunc
 		goto error;
 	}
 
-
 	if (setjmp(png_jmpbuf(png_ptr))) {
 		/* png detected error occured */
 		goto error;
@@ -258,19 +261,21 @@ gdip_load_png_image_from_file_or_stream (FILE *fp, GetBytesDelegate getBytesFunc
 		png_set_read_fn (png_ptr, getBytesFunc, _gdip_png_stream_read_data);
 	}
 
-	png_read_png(png_ptr, info_ptr, 0, NULL);
+	png_read_png (png_ptr, info_ptr, 0, NULL);
 
-	if ((png_get_bit_depth (png_ptr, info_ptr) <= 8)
-	 && (png_get_channels (png_ptr, info_ptr) == 1)
-	 && ((png_get_color_type (png_ptr, info_ptr) == PNG_COLOR_TYPE_PALETTE)
-	 || (png_get_color_type (png_ptr, info_ptr) == PNG_COLOR_TYPE_GRAY))) {
+	bit_depth = png_get_bit_depth (png_ptr, info_ptr);
+	channels = png_get_channels (png_ptr, info_ptr);
+	color_type = png_get_color_type (png_ptr, info_ptr);
+
+	/* 2bpp is a special case (promoted to 32bpp ARGB by MS GDI+) */
+	if ((bit_depth <= 8) && (bit_depth != 2) && (channels == 1) && 
+		((color_type == PNG_COLOR_TYPE_PALETTE)	|| (color_type == PNG_COLOR_TYPE_GRAY))) {
 		int		width;
 		int		height;
-		int		bit_depth;
 		int		source_stride;
 		int		dest_stride;
 		png_bytep	*row_pointers;
-		guchar		*rawptr;
+		BYTE		*rawptr;
 		int		num_colours;
 		int		palette_entries;
 		ColorPalette	*palette;
@@ -280,63 +285,27 @@ gdip_load_png_image_from_file_or_stream (FILE *fp, GetBytesDelegate getBytesFunc
 
 		width = png_get_image_width (png_ptr, info_ptr);
 		height = png_get_image_height (png_ptr, info_ptr);
-		bit_depth = png_get_bit_depth (png_ptr, info_ptr);
 
 		source_stride = (width * bit_depth + 7) / 8;
-		dest_stride = (source_stride + sizeof(pixman_bits_t) - 1) & ~(sizeof(pixman_bits_t) - 1);
+		dest_stride = source_stride;
+		gdip_align_stride (dest_stride);
 
 		/* Copy image data. */
 		row_pointers = png_get_rows (png_ptr, info_ptr);
 
-		if (bit_depth == 2) { /* upsample to 4bpp */
-			dest_stride = ((width + 1) / 2 + sizeof(pixman_bits_t) - 1) & ~(sizeof(pixman_bits_t) - 1);
-
-			rawdata = GdipAlloc(dest_stride * height);
-			for (i=0; i < height; i++) {
-				png_bytep row = row_pointers[i];
-				rawptr = rawdata + i * dest_stride;
-
-				for (j=0; j < source_stride; j++) {
-					int four_pixels = row[j];
-
-					int first_two = 0x0F & (four_pixels >> 4);
-					int second_two = 0x0F & four_pixels;
-
-					first_two = (first_two & 0x03) | ((first_two & 0x0C) << 2);
-					second_two = (second_two & 0x03) | ((second_two & 0x0C) << 2);
-
-					rawptr[j * 2 + 0] = first_two;
-					rawptr[j * 2 + 1] = second_two;
-				}
-			}
-		} else {
-			rawdata = GdipAlloc(dest_stride * height);
-			for (i=0; i < height; i++) {
-				memcpy(rawdata + i * dest_stride, row_pointers[i], source_stride);
-			}
+		rawdata = GdipAlloc(dest_stride * height);
+		for (i=0; i < height; i++) {
+			memcpy (rawdata + i * dest_stride, row_pointers[i], source_stride);
 		}
 
 		/* Copy palette. */
 		num_colours = 1 << bit_depth;
-		if (bit_depth == 4) {
-			num_colours = 256;
-		}
-
-		palette = GdipAlloc (sizeof(ColorPalette) + num_colours * sizeof(ARGB));
-
-		palette->Flags = 0;
-		palette->Count = num_colours;
 
 		if (png_get_color_type (png_ptr, info_ptr) == PNG_COLOR_TYPE_GRAY) {
 			/* A gray-scale image; generate a palette fading from black to white. */
 			colourspace_flag = ImageFlagsColorSpaceGRAY;
+			palette = gdip_create_greyscale_palette (num_colours);
 			palette->Flags = PaletteFlagsGrayScale;
-
-			for (i=0; i < num_colours; i++) {
-				int intensity = i * 255 / (num_colours - 1);
-
-				set_pixel_bgra (&palette->Entries[i], 0, intensity, intensity, intensity, 0xFF); /* alpha */
-			}
 		} else {
 			/* Copy the palette data into the GDI+ structure. */
 			colourspace_flag = ImageFlagsColorSpaceRGB;
@@ -346,6 +315,10 @@ gdip_load_png_image_from_file_or_stream (FILE *fp, GetBytesDelegate getBytesFunc
 				palette_entries = info_ptr->num_palette;
 			}
 
+			palette = GdipAlloc (sizeof(ColorPalette) + (num_colours - 1) * sizeof(ARGB));
+			palette->Flags = 0;
+			palette->Count = num_colours;
+
 			for (i=0; i < palette_entries; i++) {
 				set_pixel_bgra (&palette->Entries[i], 0,
 						info_ptr->palette[i].blue,
@@ -373,10 +346,8 @@ gdip_load_png_image_from_file_or_stream (FILE *fp, GetBytesDelegate getBytesFunc
 			}
 		}
 
-
 		result = gdip_bitmap_new_with_frame (&gdip_image_frameDimension_page_guid, TRUE);
-		result->type = imageBitmap;
-		result->cairo_format = CAIRO_FORMAT_ARGB32;
+		result->type = ImageTypeBitmap;
 		result->active_bitmap->stride = dest_stride;
 		result->active_bitmap->width = width;
 		result->active_bitmap->height = height;
@@ -384,30 +355,41 @@ gdip_load_png_image_from_file_or_stream (FILE *fp, GetBytesDelegate getBytesFunc
 		result->active_bitmap->reserved = GBD_OWN_SCAN0;
 
 		switch (bit_depth) {
-			case 1: result->active_bitmap->pixel_format = Format1bppIndexed; result->cairo_format = CAIRO_FORMAT_A1; break;
-			case 4: result->active_bitmap->pixel_format = Format4bppIndexed; result->cairo_format = CAIRO_FORMAT_A8; break;
-			case 8: result->active_bitmap->pixel_format = Format8bppIndexed; result->cairo_format = CAIRO_FORMAT_A8; break;
+		case 1:
+			result->active_bitmap->pixel_format = PixelFormat1bppIndexed;
+			result->cairo_format = CAIRO_FORMAT_A1;
+			break;
+		// note: 2bpp is a special case as the format is "promoted" to PixelFormat32bppARGB / CAIRO_FORMAT_ARGB32
+		//       we deal with this later...
+		case 4:
+			result->active_bitmap->pixel_format = PixelFormat4bppIndexed;
+			result->cairo_format = CAIRO_FORMAT_A8;
+			break;
+		case 8:
+			result->active_bitmap->pixel_format = PixelFormat8bppIndexed;
+			result->cairo_format = CAIRO_FORMAT_A8;
+			break;
 		}
 
 		result->active_bitmap->image_flags = ImageFlagsReadOnly | ImageFlagsHasRealPixelSize | colourspace_flag; /* assigned when the palette is loaded */
 		result->active_bitmap->dpi_horz = 0;
 		result->active_bitmap->dpi_vert = 0;
 		result->active_bitmap->palette = palette;
-	} else {
+	}
+
+	/* 2bpp needs to enter here too */
+	if (!result) {
 		int		width;
 		int		height;
-		guchar		bit_depth;
-		guchar		color_type;
-		int		channels;
+		BYTE		bit_depth;
 		int		stride;
 		int		interlace;
 		png_bytep *row_pointers;
-		guchar *rawptr;
+		BYTE *rawptr;
 		int i, j;
 
 		width = png_get_image_width (png_ptr, info_ptr);
 		height = png_get_image_height (png_ptr, info_ptr);
-		channels = png_get_channels (png_ptr, info_ptr);
 
 		/* 24 and 32bpp are supported, 48, 64bpp aren't, see http://bugzilla.ximian.com/show_bug.cgi?id=80693 */
 		bit_depth = png_get_bit_depth (png_ptr, info_ptr);
@@ -417,7 +399,6 @@ gdip_load_png_image_from_file_or_stream (FILE *fp, GetBytesDelegate getBytesFunc
 			goto error;
 		}
 
-		color_type = png_get_color_type (png_ptr, info_ptr);
 		interlace = png_get_interlace_type (png_ptr, info_ptr);
 
 		/* According to the libpng manual, this sequence is equivalent to
@@ -434,10 +415,11 @@ gdip_load_png_image_from_file_or_stream (FILE *fp, GetBytesDelegate getBytesFunc
 			png_set_tRNS_to_alpha(png_ptr);
 		}
 
-		stride = (width * 4) + (sizeof(pixman_bits_t)-1);
-		stride &= ~(sizeof(pixman_bits_t)-1);
+		stride = (width * 4);
+		gdip_align_stride (stride);
 
 		row_pointers = png_get_rows (png_ptr, info_ptr);
+
 		rawdata = GdipAlloc (stride * height);
 		rawptr = rawdata;
 
@@ -446,13 +428,13 @@ gdip_load_png_image_from_file_or_stream (FILE *fp, GetBytesDelegate getBytesFunc
 				for (i = 0; i < height; i++) {
 					png_bytep rowp = row_pointers[i];
 					for (j = 0; j < width; j++) {
-						byte a = rowp[3];
+						BYTE a = rowp[3];
 						if (a == 0) {
 							set_pixel_bgra (rawptr, 0, 0, 0, 0, 0);
 						} else {
-							byte b = rowp[2];
-							byte g = rowp[1];
-							byte r = rowp[0];
+							BYTE b = rowp[2];
+							BYTE g = rowp[1];
+							BYTE r = rowp[0];
 
 							if (a < 0xff) {
 								r = pre_multiplied_table [r][a];
@@ -484,39 +466,78 @@ gdip_load_png_image_from_file_or_stream (FILE *fp, GetBytesDelegate getBytesFunc
 			case 1:
 				for (i = 0; i < height; i++) {
 					png_bytep rowp = row_pointers[i];
-					for (j = 0; j < width; j++) {
-						png_byte pix = *rowp++;
-						set_pixel_bgra (rawptr, 0, pix, pix, pix, 0xff);
-						rawptr += 4;
+					if (bit_depth == 2) {
+						// 4 pixels for each byte
+						for (j = 0; j < (width >> bit_depth); j++) {
+							png_byte palette = 0;
+							png_byte pix = *rowp++;
+
+							palette = (pix >> 6) & 0x03;
+							set_pixel_bgra (rawptr, 0,
+								info_ptr->palette[palette].blue,
+								info_ptr->palette[palette].green,
+								info_ptr->palette[palette].red,
+								0xFF); /* alpha */
+
+							palette = (pix >> 4) & 0x03;
+							set_pixel_bgra (rawptr, 4,
+								info_ptr->palette[palette].blue,
+								info_ptr->palette[palette].green,
+								info_ptr->palette[palette].red,
+								0xFF); /* alpha */
+
+							palette = (pix >> 2) & 0x03;
+							set_pixel_bgra (rawptr, 8,
+								info_ptr->palette[palette].blue,
+								info_ptr->palette[palette].green,
+								info_ptr->palette[palette].red,
+								0xFF); /* alpha */
+
+							palette = pix & 0x03;
+							set_pixel_bgra (rawptr, 12,
+								info_ptr->palette[palette].blue,
+								info_ptr->palette[palette].green,
+								info_ptr->palette[palette].red,
+								0xFF); /* alpha */
+							rawptr += 16;
+						}
+					} else {
+						for (j = 0; j < width; j++) {
+							png_byte pix = *rowp++;
+							set_pixel_bgra (rawptr, 0, pix, pix, pix, 0xff);
+							rawptr += 4;
+						}
 					}
 				}
 				break;
 		}
 
 		result = gdip_bitmap_new_with_frame (&gdip_image_frameDimension_page_guid, TRUE);
-		result->type = imageBitmap;
+		result->type = ImageTypeBitmap;
 
 		result->cairo_format = CAIRO_FORMAT_ARGB32;
 		result->active_bitmap->stride = stride;
-		result->active_bitmap->pixel_format = Format32bppArgb;
+		result->active_bitmap->pixel_format = PixelFormat32bppARGB;
 		result->active_bitmap->width = width;
 		result->active_bitmap->height = height;
 		result->active_bitmap->scan0 = rawdata;
 		result->active_bitmap->reserved = GBD_OWN_SCAN0;
 
-		result->surface = cairo_image_surface_create_for_data ((unsigned char *)rawdata,
+		result->surface = cairo_image_surface_create_for_data ((BYTE*)rawdata,
 			result->cairo_format,
 			result->active_bitmap->width,
 			result->active_bitmap->height,
 			result->active_bitmap->stride);
+
 		if (channels == 3) {
-			result->active_bitmap->pixel_format = Format24bppRgb;
+			result->active_bitmap->pixel_format = PixelFormat24bppRGB;
 			result->active_bitmap->image_flags = ImageFlagsColorSpaceRGB;
 		} else if (channels == 4) {
-			result->active_bitmap->pixel_format = Format32bppArgb;
+			result->active_bitmap->pixel_format = PixelFormat32bppARGB;
 			result->active_bitmap->image_flags = ImageFlagsColorSpaceRGB;
-		} else if (channels == 1) {
-			result->active_bitmap->pixel_format = Format8bppIndexed;
+		} else if ((channels == 1) && (color_type == PNG_COLOR_TYPE_GRAY)) {
+			// doesn't apply to 2bpp images
+			result->active_bitmap->pixel_format = PixelFormat8bppIndexed;
 			result->active_bitmap->image_flags = ImageFlagsColorSpaceGRAY;
 		}
 
@@ -593,41 +614,41 @@ gdip_save_png_image_to_file_or_stream (FILE *fp, PutBytesDelegate putBytesFunc,
 	}
 
 	switch (image->active_bitmap->pixel_format) {
-		case Format32bppArgb:
-		case Format32bppPArgb:
-		case Format32bppRgb:
+		case PixelFormat32bppARGB:
+		case PixelFormat32bppPARGB:
+		case PixelFormat32bppRGB:
 			color_type = PNG_COLOR_TYPE_RGB_ALPHA;
 			bit_depth = 8;
 			break;
 
-		case Format24bppRgb:
+		case PixelFormat24bppRGB:
 			color_type = PNG_COLOR_TYPE_RGB; /* FIXME - we should be able to write grayscale PNGs */
 			bit_depth = 8;
 			break;
 
-		case Format8bppIndexed:
+		case PixelFormat8bppIndexed:
 			color_type = PNG_COLOR_TYPE_PALETTE;
 			bit_depth = 8;
 			break;
 
-		case Format4bppIndexed:
+		case PixelFormat4bppIndexed:
 			color_type = PNG_COLOR_TYPE_PALETTE;
 			bit_depth = 4;
 			break;
 
-		case Format1bppIndexed:
+		case PixelFormat1bppIndexed:
 			color_type = PNG_COLOR_TYPE_PALETTE;
 			bit_depth = 1;
 			break;
 
 		/* We're not going to even try to save these images, for now */
-		case Format64bppArgb:
-		case Format64bppPArgb:
-		case Format48bppRgb:
-		case Format16bppArgb1555:
-		case Format16bppGrayScale:
-		case Format16bppRgb555:
-		case Format16bppRgb565:
+		case PixelFormat64bppARGB:
+		case PixelFormat64bppPARGB:
+		case PixelFormat48bppRGB:
+		case PixelFormat16bppARGB1555:
+		case PixelFormat16bppGrayScale:
+		case PixelFormat16bppRGB555:
+		case PixelFormat16bppRGB565:
 		default:
 			color_type = -1;
 			bit_depth = -1;
@@ -645,7 +666,7 @@ gdip_save_png_image_to_file_or_stream (FILE *fp, PutBytesDelegate putBytesFunc,
 		png_color palette[256];
 
 		int palette_entries = image->active_bitmap->palette->Count;
-		if (image->active_bitmap->pixel_format == Format4bppIndexed) {
+		if (image->active_bitmap->pixel_format == PixelFormat4bppIndexed) {
 			palette_entries = 16;
 		}
 
@@ -665,7 +686,7 @@ gdip_save_png_image_to_file_or_stream (FILE *fp, PutBytesDelegate putBytesFunc,
 	png_write_info (png_ptr, info_ptr);
 
 #if 0
-	if ((image->active_bitmap->pixel_format == Format24bppRgb) || (image->active_bitmap->pixel_format == Format32bppRgb)) {
+	if ((image->active_bitmap->pixel_format == Format24bppRGB) || (image->active_bitmap->pixel_format == Format32bppRGB)) {
 		png_set_filler (png_ptr, 0, PNG_FILLER_AFTER);
 	} else if (image->active_bitmap->pixel_format == Format8bppIndexed) {
 		png_set_filler (png_ptr, 0, PNG_FILLER_AFTER);
@@ -678,19 +699,19 @@ gdip_save_png_image_to_file_or_stream (FILE *fp, PutBytesDelegate putBytesFunc,
 		for (i = 0; i < image->active_bitmap->height; i++) {
 			png_write_row (png_ptr, image->active_bitmap->scan0 + i * image->active_bitmap->stride);
 		}
-	} else if (image->active_bitmap->pixel_format == Format24bppRgb) {
+	} else if (image->active_bitmap->pixel_format == PixelFormat24bppRGB) {
 		int j;
-		guchar *row_pointer = GdipAlloc (image->active_bitmap->width * 3);
+		BYTE *row_pointer = GdipAlloc (image->active_bitmap->width * 3);
 		for (i = 0; i < image->active_bitmap->height; i++) {
 			for (j = 0; j < image->active_bitmap->width; j++) {
 #ifdef WORDS_BIGENDIAN
-				row_pointer[j*3  ] = *((guchar *)image->active_bitmap->scan0 + (image->active_bitmap->stride * i) + (j*4) + 3);
-				row_pointer[j*3+1] = *((guchar *)image->active_bitmap->scan0 + (image->active_bitmap->stride * i) + (j*4) + 2);
-				row_pointer[j*3+2] = *((guchar *)image->active_bitmap->scan0 + (image->active_bitmap->stride * i) + (j*4) + 1);
+				row_pointer[j*3  ] = *((BYTE*)image->active_bitmap->scan0 + (image->active_bitmap->stride * i) + (j*4) + 3);
+				row_pointer[j*3+1] = *((BYTE*)image->active_bitmap->scan0 + (image->active_bitmap->stride * i) + (j*4) + 2);
+				row_pointer[j*3+2] = *((BYTE*)image->active_bitmap->scan0 + (image->active_bitmap->stride * i) + (j*4) + 1);
 #else
-				row_pointer[j*3  ] = *((guchar *)image->active_bitmap->scan0 + (image->active_bitmap->stride * i) + (j*4) + 0);
-				row_pointer[j*3+1] = *((guchar *)image->active_bitmap->scan0 + (image->active_bitmap->stride * i) + (j*4) + 1);
-				row_pointer[j*3+2] = *((guchar *)image->active_bitmap->scan0 + (image->active_bitmap->stride * i) + (j*4) + 2);
+				row_pointer[j*3  ] = *((BYTE*)image->active_bitmap->scan0 + (image->active_bitmap->stride * i) + (j*4) + 0);
+				row_pointer[j*3+1] = *((BYTE*)image->active_bitmap->scan0 + (image->active_bitmap->stride * i) + (j*4) + 1);
+				row_pointer[j*3+2] = *((BYTE*)image->active_bitmap->scan0 + (image->active_bitmap->stride * i) + (j*4) + 2);
 #endif /* WORDS_BIGENDIAN */
 			}
 			png_write_row (png_ptr, row_pointer);
@@ -699,14 +720,14 @@ gdip_save_png_image_to_file_or_stream (FILE *fp, PutBytesDelegate putBytesFunc,
 	} else {
 #ifdef WORDS_BIGENDIAN
 		int j;
-		guchar *row_pointer = GdipAlloc (image->active_bitmap->width * 4);
+		BYTE *row_pointer = GdipAlloc (image->active_bitmap->width * 4);
 
 		for (i = 0; i < image->active_bitmap->height; i++) {
 			for (j = 0; j < image->active_bitmap->width; j++) {
-				row_pointer[j*4] = *((guchar *)image->active_bitmap->scan0 + (image->active_bitmap->stride * i) + (j*4) + 3);
-				row_pointer[j*4+1] = *((guchar *)image->active_bitmap->scan0 + (image->active_bitmap->stride * i) + (j*4) + 2);
-				row_pointer[j*4+2] = *((guchar *)image->active_bitmap->scan0 + (image->active_bitmap->stride * i) + (j*4) + 1);
-				row_pointer[j*4+3] = *((guchar *)image->active_bitmap->scan0 + (image->active_bitmap->stride * i) + (j*4) + 0);
+				row_pointer[j*4]   = *((BYTE*)image->active_bitmap->scan0 + (image->active_bitmap->stride * i) + (j*4) + 3);
+				row_pointer[j*4+1] = *((BYTE*)image->active_bitmap->scan0 + (image->active_bitmap->stride * i) + (j*4) + 2);
+				row_pointer[j*4+2] = *((BYTE*)image->active_bitmap->scan0 + (image->active_bitmap->stride * i) + (j*4) + 1);
+				row_pointer[j*4+3] = *((BYTE*)image->active_bitmap->scan0 + (image->active_bitmap->stride * i) + (j*4) + 0);
 			}
 			png_write_row (png_ptr, row_pointer);
 		}
