@@ -1,6 +1,6 @@
 /*
  * Copyright (c) 2004-2005 Ximian
- * Copyright (C) 2006 Novell, Inc (http://www.novell.com)
+ * Copyright (C) 2006-2007 Novell, Inc (http://www.novell.com)
  *
  * Permission is hereby granted, free of charge, to any person obtaining a copy of this software
  * and associated documentation files (the "Software"), to deal in the Software without restriction,
@@ -23,14 +23,14 @@
  *
  */
 
-#include "region.h"
-#include "matrix.h"
+#include "region-private.h"
+#include "graphics-path-private.h"
 
 /*
 	Helper functions
 */
 
-void
+static void
 gdip_from_Rect_To_RectF (GpRect* rect, GpRectF* rectf)
 {
         rectf->X = rect->X;
@@ -39,7 +39,7 @@ gdip_from_Rect_To_RectF (GpRect* rect, GpRectF* rectf)
         rectf->Height = rect->Height;
 }
 
-void
+static void
 gdip_add_rect_to_array (GpRectF** srcarray, int* elements,  GpRectF* rect)
 {
         GpRectF *array, *next;
@@ -58,7 +58,7 @@ gdip_add_rect_to_array (GpRectF** srcarray, int* elements,  GpRectF* rect)
         *elements = *elements + 1;
 }
 
-BOOL
+static BOOL
 gdip_is_Point_in_RectF_Visible (float x, float y, GpRectF* rect)
 {
         if ((x >= rect->X && x < (rect->X + rect->Width))
@@ -68,7 +68,7 @@ gdip_is_Point_in_RectF_Visible (float x, float y, GpRectF* rect)
                 return FALSE;
 }
 
-BOOL
+static BOOL
 gdip_is_Point_in_RectFs_Visible (float x, float y, GpRectF* r, int cnt)
 {
         GpRectF* rect = r;
@@ -83,7 +83,40 @@ gdip_is_Point_in_RectFs_Visible (float x, float y, GpRectF* r, int cnt)
         return FALSE;
 }
 
-void gdip_get_bounds (GpRectF *allrects, int allcnt, GpRectF *bound);
+static void
+gdip_get_bounds (GpRectF *allrects, int allcnt, GpRectF *bound)
+{
+        float nx, ny, fx, fy;
+        int i;
+        GpRectF *rect;
+
+        if (allrects == NULL || allcnt == 0) {
+                bound->X = bound->Y = bound->Width =  bound->Height = 0;
+                return;
+        }
+
+        /* Build a rect that contains all the rects inside. Smallest x,y and biggest x,y*/
+        nx = allrects->X; ny = allrects->Y;
+        fx = allrects->X + allrects->Width; fy = allrects->Y + allrects->Height;
+
+        for (i = 0, rect = allrects; i < allcnt; i++, rect++) {
+
+                if (rect->X < nx)
+                        nx = rect->X;
+
+                if (rect->Y < ny)
+                        ny = rect->Y;
+
+                if (rect->X + rect->Width  > fx)
+                        fx = rect->X + rect->Width;
+
+                if (rect->Y + rect->Height > fy)
+                        fy = rect->Y + rect->Height;
+        }
+
+        bound->X = nx; bound->Y = ny;
+        bound->Width = fx - nx; bound->Height = fy - ny;
+}
 
 /* This internal version doesn't require a Graphic object to work */
 static BOOL
@@ -113,7 +146,7 @@ gdip_is_region_empty (GpRegion *region)
 static BOOL
 gdip_is_rect_infinite (GpRectF *rect)
 {
-	return ((rect->X == REGION_INFINITE_POSITION) && 
+	return (rect && (rect->X == REGION_INFINITE_POSITION) && 
 		(rect->Y == REGION_INFINITE_POSITION) &&
 		(rect->Width == REGION_INFINITE_LENGTH) && 
 		(rect->Height == REGION_INFINITE_LENGTH));
@@ -142,7 +175,7 @@ gdip_is_InfiniteRegion (GpRegion *region)
 	return FALSE;
 }
 
-BOOL
+static BOOL
 gdip_intersects (GpRectF *rect1, GpRectF *rect2)
 {
 	if (rect1->X + rect1->Width == rect2->X) {
@@ -156,7 +189,7 @@ gdip_intersects (GpRectF *rect1, GpRectF *rect2)
 }
 
 /* Is source contained in target ? */
-BOOL
+static BOOL
 gdip_contains (GpRectF *rect1, GpRectF *rect2)
 {
 	return (rect1->X >= rect2->X &&
@@ -165,7 +198,7 @@ gdip_contains (GpRectF *rect1, GpRectF *rect2)
 		rect1->Y + rect1->Height <= rect2->Y + rect2->Height);
 }
 
-BOOL
+static BOOL
 gdip_add_rect_to_array_notcontained (GpRectF** srcarray, int* elements,  GpRectF* rect)
 {
 	int i;
@@ -185,7 +218,7 @@ gdip_add_rect_to_array_notcontained (GpRectF** srcarray, int* elements,  GpRectF
 }
 
 
-BOOL
+static BOOL
 gdip_equals (GpRectF *rect1, GpRectF *rect2)
 {
 	return (rect1->X == rect2->X &&
@@ -205,7 +238,7 @@ gdip_is_Point_in_RectF_inclusive (float x, float y, GpRectF* rect)
 }
 
 /* Finds a rect that has the lowest x and y after the src rect provided */
-BOOL
+static BOOL
 gdip_getlowestrect (GpRectF *rects, int cnt, GpRectF* src, GpRectF* rslt)
 {
 	int i;
@@ -328,7 +361,7 @@ gdip_region_create_from_path (GpRegion *region, GpPath *path)
 	return GdipClonePath (path, &region->tree->path);
 }
 
-GpStatus
+static GpStatus
 gdip_createRegion (GpRegion **region, RegionType type, void *src)
 {
         GpRegion *result;
@@ -519,44 +552,8 @@ GdipSetEmpty (GpRegion *region)
         return Ok;
 }
 
-void
-gdip_get_bounds (GpRectF *allrects, int allcnt, GpRectF *bound)
-{
-        float nx, ny, fx, fy;
-        int i;
-        GpRectF *rect;
-
-        if (allrects == NULL || allcnt == 0) {
-                bound->X = bound->Y = bound->Width =  bound->Height = 0;
-                return;
-        }
-
-        /* Build a rect that contains all the rects inside. Smallest x,y and biggest x,y*/
-        nx = allrects->X; ny = allrects->Y;
-        fx = allrects->X + allrects->Width; fy = allrects->Y + allrects->Height;
-
-        for (i = 0, rect = allrects; i < allcnt; i++, rect++) {
-
-                if (rect->X < nx)
-                        nx = rect->X;
-
-                if (rect->Y < ny)
-                        ny = rect->Y;
-
-                if (rect->X + rect->Width  > fx)
-                        fx = rect->X + rect->Width;
-
-                if (rect->Y + rect->Height > fy)
-                        fy = rect->Y + rect->Height;
-        }
-
-        bound->X = nx; bound->Y = ny;
-        bound->Width = fx - nx; bound->Height = fy - ny;
-}
-
-
 /* Exclude */
-void
+static void
 gdip_combine_exclude (GpRegion *region, GpRectF *rtrg, int cntt)
 {
 	GpRectF *allsrcrects = NULL, *rects = NULL;
@@ -677,7 +674,7 @@ gdip_combine_exclude (GpRegion *region, GpRectF *rtrg, int cntt)
 	Complement: the part of the second region not shared with the first region.
 	Scans the region to be combined and store the rects not present in the region
 */
-void
+static void
 gdip_combine_complement (GpRegion *region, GpRectF *rtrg, int cntt)
 {
 	GpRegion regsrc;
@@ -707,7 +704,7 @@ gdip_combine_complement (GpRegion *region, GpRectF *rtrg, int cntt)
 
 
 /* Union */
-void
+static void
 gdip_combine_union (GpRegion *region, GpRectF *rtrg, int cnttrg)
 {
 	GpRectF *allrects = NULL, *rects = NULL;
@@ -850,7 +847,7 @@ gdip_combine_union (GpRegion *region, GpRectF *rtrg, int cnttrg)
 }
 
 /* Intersect */
-void
+static void
 gdip_combine_intersect (GpRegion *region, GpRectF *rtrg, int cnttrg)
 {
 	GpRectF *rectsrc;
@@ -893,7 +890,7 @@ gdip_combine_intersect (GpRegion *region, GpRectF *rtrg, int cnttrg)
 }
 
 /* Xor */
-void
+static void
 gdip_combine_xor (GpRegion *region, GpRectF *recttrg, int cnttrg)
 {
         GpRegion *rgnsrc;  /* All rectangles of both regions*/
@@ -1037,6 +1034,9 @@ GdipCombineRegionPath (GpRegion *region, GpPath *path, CombineMode combineMode)
 
 	/* special case #3 - the region is infinite */
 	if (gdip_is_InfiniteRegion (region)) {
+		/* additional shortcuts are possible if path is empty */
+		BOOL empty = (path->count == 0);
+
 		switch (combineMode) {
 		case CombineModeUnion:
 			/* Union with infinity is a no-op (still an infinite region) */
@@ -1049,10 +1049,15 @@ GdipCombineRegionPath (GpRegion *region, GpPath *path, CombineMode combineMode)
 		case CombineModeIntersect:
 			/* Intersection with infinity is the path itself */
 			gdip_clear_region (region);
-			gdip_region_create_from_path (region, path);
+			if (empty)
+				region->type = RegionTypeRectF;
+			else
+				gdip_region_create_from_path (region, path);
 			return Ok;
 		default:
-			/* Xor and Exclude must be treated as a "normal" case */
+			/* Xor and Exclude must be treated as a "normal" case unless the path is empty */
+			if (empty)
+				return Ok;
 			break;
 		}
 	}
@@ -1163,6 +1168,9 @@ GdipCombineRegionRegion (GpRegion *region,  GpRegion *region2, CombineMode combi
 			break;
 		}
 	} else if (gdip_is_InfiniteRegion (region)) {
+		/* additional shortcuts are possible if path is empty */
+		BOOL empty = gdip_is_region_empty (region2);
+
 		switch (combineMode) {
 		case CombineModeUnion:
 			/* Union with infinity is a no-op (still an infinite region) */
@@ -1176,11 +1184,14 @@ GdipCombineRegionRegion (GpRegion *region,  GpRegion *region2, CombineMode combi
 			/* Intersection with infinity is the path itself (like an Union with Empty) */
 			gdip_clear_region (region);
 			region->type = RegionTypeRectF;
+			if (empty)
+				return Ok;
 			combineMode = CombineModeUnion; 
 			break;
 		default:
-			/* Xor and Exclude must be treated as a "normal" case */
-			break;
+			/* Xor must be treated as a "normal" case unless the path is empty */
+			if (empty)
+				return Ok;
 		}
 	}
 
@@ -1523,6 +1534,12 @@ GdipTranslateRegion (GpRegion *region, float dx, float dy)
         if (!region)
                 return InvalidParameter;
 
+	/* can't transforman infinite region to anything else than an infinite region 
+	 * (even if you scale it by half it's still infinite ;-) see unit tests
+	 */
+	if (gdip_is_InfiniteRegion (region))
+		return Ok;
+
 	if (region->type == RegionTypePath) {
 		gdip_region_translate_tree (region->tree, dx, dy);
 		/* any existing bitmap is still valid _if_ we update it's origin */
@@ -1548,10 +1565,31 @@ GdipTranslateRegionI (GpRegion *region, int dx, int dy)
         return GdipTranslateRegion (region, dx, dy);
 }
 
+/* this call doesn't exists in GDI+ */
+static GpStatus
+ScaleRegion (GpRegion *region, float sx, float sy)
+{
+	if (!region)
+		return InvalidParameter;
+
+	if ((region->type == RegionTypeRectF) && region->rects) {
+	        int i;
+	        GpRectF *rect;
+                for (i = 0, rect = region->rects ; i < region->cnt; i++, rect++) {
+                        rect->X *= sx;
+                        rect->Y *= sy;
+                        rect->Width *= sx;
+                        rect->Height *= sy;
+                }
+        }
+
+        return Ok;
+}
+
 GpStatus
 GdipTransformRegion (GpRegion *region, GpMatrix *matrix)
 {
-	GpStatus status;
+	GpStatus status = Ok;
 
 	if (!region || !matrix)
 		return InvalidParameter;
@@ -1564,12 +1602,35 @@ GdipTransformRegion (GpRegion *region, GpMatrix *matrix)
 	if (gdip_is_matrix_empty (matrix))
 		return Ok;
 
-	/* avoid heavy stuff (e.g. conversion to path, invalidating bitmap...)
-	   if the transform is only to do a simple translation */
-	if (gdip_is_matrix_a_translation (matrix)) {
-		return GdipTranslateRegion (region, 
-			gdip_matrix_get_x_translation (matrix), 
-			gdip_matrix_get_y_translation (matrix));
+	/* can't transforman infinite region to anything else than an infinite region 
+	 * (even if you scale it by half it's still infinite ;-) see unit tests
+	 */
+	if (gdip_is_InfiniteRegion (region))
+		return Ok;
+
+	/* try to avoid heavy stuff (e.g. conversion to path, invalidating 
+	 * bitmap...) if the transform is:
+	 * - a translation + scale operations (for rectangle ebased region)
+	 * - only to do a scale operation (for a rectangle based region)
+	 * - only to do a simple translation (for both rectangular and bitmap based regions)
+	 */
+	if ((matrix->xy == 0.0f) && (matrix->yx == 0.0f)) {
+		BOOL s = (((matrix->xx != 1.0f) || (matrix->yy != 1.0f)) && (region->type == RegionTypeRectF));
+		BOOL t = ((matrix->x0 != 0.0f) || (matrix->yx != 0.0f));
+		if (s) {
+			status = ScaleRegion (region, 
+				gdip_matrix_get_x_scale (matrix), 
+				gdip_matrix_get_y_scale (matrix));
+		}
+		if (t && (status == Ok)) {
+			status = GdipTranslateRegion (region, 
+				gdip_matrix_get_x_translation (matrix), 
+				gdip_matrix_get_y_translation (matrix));
+		}
+
+		/* return now if we could optimize the transform (to avoid bitmaps) */
+		if (t || s)
+			return status;
 	}
 
 	/* most matrix operations would change the rectangles into path so we always preempt this */
@@ -1709,11 +1770,14 @@ GdipGetRegionHRgn (GpRegion *region, GpGraphics *graphics, HRGN *hRgn)
 		return InvalidParameter;
 
 	/* infinite region returns NULL */
-	if (gdip_is_InfiniteRegion (region))
+	if (gdip_is_InfiniteRegion (region)) {
 		*hRgn = NULL;
-	else
-		*hRgn = region;
-	return Ok;
+		return Ok;
+	}
+
+	/* calling GdipGetRegionHRgn multiple times returns a different HRNG value
+	   (i.e. each to be freed separately) */
+	return GdipCloneRegion (region, (GpRegion**)hRgn);
 }
 
 // coverity[+alloc : arg-*1]
